import pytest

from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.state.base import BaseState


class TestState(BaseState):
    pass


def test_concat_expression_happy_path():
    # GIVEN two lists
    state = TestState()
    lhs_ref = ConstantValueReference([1, 2, 3])
    rhs_ref = ConstantValueReference([4, 5, 6])
    concat_expr = lhs_ref.concat(rhs_ref)

    # WHEN we resolve the expression
    result = concat_expr.resolve(state)

    # THEN the lists should be concatenated
    assert result == [1, 2, 3, 4, 5, 6]


def test_concat_expression_lhs_fail():
    # GIVEN a non-list lhs and a list rhs
    state = TestState()
    lhs_ref = ConstantValueReference(0)
    rhs_ref = ConstantValueReference([4, 5, 6])
    concat_expr = lhs_ref.concat(rhs_ref)

    # WHEN we attempt to resolve the expression
    with pytest.raises(InvalidExpressionException) as exc_info:
        concat_expr.resolve(state)

    # THEN an exception should be raised
    assert "Expected LHS to be a list, got <class 'int'>" in str(exc_info.value)


def test_concat_expression_rhs_fail():
    # GIVEN a list lhs and a non-list rhs
    state = TestState()
    lhs_ref = ConstantValueReference([1, 2, 3])
    rhs_ref = ConstantValueReference(False)
    concat_expr = lhs_ref.concat(rhs_ref)

    # WHEN we attempt to resolve the expression
    with pytest.raises(InvalidExpressionException) as exc_info:
        concat_expr.resolve(state)

    # THEN an exception should be raised
    assert "Expected RHS to be a list, got <class 'bool'>" in str(exc_info.value)



# This file was auto-generated by Fern from our API Definition.

import typing

from .scenario_input_chat_history_variable_value import ScenarioInputChatHistoryVariableValue
from .scenario_input_json_variable_value import ScenarioInputJsonVariableValue
from .scenario_input_string_variable_value import ScenarioInputStringVariableValue

ScenarioInput = typing.Union[
    ScenarioInputStringVariableValue, ScenarioInputJsonVariableValue, ScenarioInputChatHistoryVariableValue
]



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_merge_node.await_all_workflow import AwaitAllPassingWorkflow


def test_serialize_workflow__await_all():
    # GIVEN a Workflow that uses an await all merge node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=AwaitAllPassingWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {"id": "959ba00d-d30b-402e-8676-76efc4c3d2ae", "key": "value", "type": "STRING"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 6
    assert len(workflow_raw_data["nodes"]) == 6

    # AND each node should be serialized correctly
    entrypoint_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "ENTRYPOINT")
    assert entrypoint_node == {
        "id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    passthrough_nodes = [node for node in workflow_raw_data["nodes"] if node["type"] == "GENERIC"]
    assert len(passthrough_nodes) == 3

    merge_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "MERGE")
    assert not DeepDiff(
        {
            "id": "37c10e8a-771b-432b-a767-31f5007851f0",
            "type": "MERGE",
            "inputs": [],
            "data": {
                "label": "Await All Merge Node",
                "merge_strategy": "AWAIT_ALL",
                "target_handles": [
                    {"id": "f40ff7fb-de1b-4aa4-ba3c-7630f7357cbf"},
                    {"id": "42eeb66c-9792-4609-8c71-3a56f668f4dc"},
                ],
                "source_handle_id": "3bbc469f-0fb0-4b3d-a28b-746fefec2818",
            },
            "display_data": {"position": {"x": 400.0, "y": -50.0}},
            "base": {
                "module": ["vellum", "workflows", "nodes", "displayable", "merge_node", "node"],
                "name": "MergeNode",
            },
            "definition": {
                "module": ["tests", "workflows", "basic_merge_node", "await_all_workflow"],
                "name": "AwaitAllMergeNode",
            },
            "trigger": {
                "id": "0efd256f-f5f6-45fe-9adb-651780f5e63d",
                "merge_behavior": "AWAIT_ALL",
            },
            "ports": [{"id": "3bbc469f-0fb0-4b3d-a28b-746fefec2818", "name": "default", "type": "DEFAULT"}],
        },
        merge_node,
        ignore_order_func=lambda x: x.path() == "root['data']['target_handles']",
    )

    final_output_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "TERMINAL")
    assert final_output_node == {
        "id": "8187ce10-62b7-4a2c-8c0f-297387915467",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "value",
            "target_handle_id": "ff55701c-16d3-4348-a633-6a298e71377d",
            "output_id": "959ba00d-d30b-402e-8676-76efc4c3d2ae",
            "output_type": "STRING",
            "node_input_id": "fed9d343-6504-460c-968b-d3f9658193d0",
        },
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "final_output_node",
                "node",
            ],
            "name": "FinalOutputNode",
        },
        "definition": None,
        "inputs": [
            {
                "id": "fed9d343-6504-460c-968b-d3f9658193d0",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "634f0202-9ea9-4c62-b152-1a58c595cffb",
                                "output_id": "d4266640-9718-4a74-b24b-500448d87871",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 800.0, "y": -50.0}},
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "9a65dd52-c3eb-496e-9d34-46b39534a261",
                "source_node_id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
                "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
                "target_node_id": "59243c65-053f-4ea6-9157-3f3edb1477bf",
                "target_handle_id": "e622fe61-3bca-4aff-86e1-25dad7bdf9d4",
                "type": "DEFAULT",
            },
            {
                "id": "e5598f3c-fb00-4f25-a0a6-9fb6af9d69a8",
                "source_node_id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
                "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
                "target_node_id": "127ef456-91bc-43c6-bd8b-1772db5e3cb5",
                "target_handle_id": "e5cc41cb-71db-43ec-b3f0-c78706af3351",
                "type": "DEFAULT",
            },
            {
                "id": "8ff20817-974e-4a3a-bb65-f0ad73557649",
                "source_node_id": "59243c65-053f-4ea6-9157-3f3edb1477bf",
                "source_handle_id": "b9c5f52b-b714-46e8-a09c-38b4e770dd36",
                "target_node_id": "37c10e8a-771b-432b-a767-31f5007851f0",
                "target_handle_id": "42eeb66c-9792-4609-8c71-3a56f668f4dc",
                "type": "DEFAULT",
            },
            {
                "id": "0d8c801c-d76a-437a-831a-530885b75f96",
                "source_node_id": "127ef456-91bc-43c6-bd8b-1772db5e3cb5",
                "source_handle_id": "b0bd17f3-4ce6-4232-9666-ec8afa161bf2",
                "target_node_id": "37c10e8a-771b-432b-a767-31f5007851f0",
                "target_handle_id": "f40ff7fb-de1b-4aa4-ba3c-7630f7357cbf",
                "type": "DEFAULT",
            },
            {
                "id": "70c1005d-339a-41bc-b6c2-10bc30a0281c",
                "source_node_id": "37c10e8a-771b-432b-a767-31f5007851f0",
                "source_handle_id": "3bbc469f-0fb0-4b3d-a28b-746fefec2818",
                "target_node_id": "634f0202-9ea9-4c62-b152-1a58c595cffb",
                "target_handle_id": "acd48f48-54fb-4b2b-ab37-96d336f6dfb3",
                "type": "DEFAULT",
            },
            {
                "id": "3d031c93-09b1-4937-9f98-c30a7ba6823d",
                "source_node_id": "634f0202-9ea9-4c62-b152-1a58c595cffb",
                "source_handle_id": "de32167b-cf53-4df5-a344-1b9be852e9ff",
                "target_node_id": "8187ce10-62b7-4a2c-8c0f-297387915467",
                "target_handle_id": "ff55701c-16d3-4348-a633-6a298e71377d",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "AwaitAllPassingWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_merge_node",
            "await_all_workflow",
        ],
    }



from typing import TYPE_CHECKING, Any, Dict, Generic, List, Literal, Optional, Set, Type, Union

from pydantic import field_serializer, model_serializer

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.workflows.errors import WorkflowError
from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.ports.port import Port
from vellum.workflows.references.node import NodeReference
from vellum.workflows.types.definition import serialize_type_encoder_with_id
from vellum.workflows.types.generics import OutputsType

from .types import BaseEvent, default_serializer

if TYPE_CHECKING:
    from vellum.workflows.nodes.bases import BaseNode


class _BaseNodeExecutionBody(UniversalBaseModel):
    node_definition: Type["BaseNode"]

    @field_serializer("node_definition")
    def serialize_node_definition(self, node_definition: Type, _info: Any) -> Dict[str, Any]:
        return serialize_type_encoder_with_id(node_definition)


class _BaseNodeEvent(BaseEvent):
    body: _BaseNodeExecutionBody

    @property
    def node_definition(self) -> Type["BaseNode"]:
        return self.body.node_definition


NodeInputName = Union[NodeReference, AccessorExpression]
InvokedPorts = Optional[Set["Port"]]


class NodeExecutionInitiatedBody(_BaseNodeExecutionBody):
    inputs: Dict[NodeInputName, Any]

    @field_serializer("inputs")
    def serialize_inputs(self, inputs: Dict[NodeInputName, Any], _info: Any) -> Dict[str, Any]:
        return default_serializer({descriptor.name: value for descriptor, value in inputs.items()})


class NodeExecutionInitiatedEvent(_BaseNodeEvent):
    name: Literal["node.execution.initiated"] = "node.execution.initiated"
    body: NodeExecutionInitiatedBody

    @property
    def inputs(self) -> Dict[NodeInputName, Any]:
        return self.body.inputs


class NodeExecutionStreamingBody(_BaseNodeExecutionBody):
    output: BaseOutput
    invoked_ports: InvokedPorts = None

    @field_serializer("output")
    def serialize_output(self, output: BaseOutput, _info: Any) -> Dict[str, Any]:
        return default_serializer(output)

    @field_serializer("invoked_ports")
    def serialize_invoked_ports(self, invoked_ports: InvokedPorts, _info: Any) -> Optional[List[Dict[str, Any]]]:
        if not invoked_ports:
            return None
        return [default_serializer(port) for port in invoked_ports]


class NodeExecutionStreamingEvent(_BaseNodeEvent):
    name: Literal["node.execution.streaming"] = "node.execution.streaming"
    body: NodeExecutionStreamingBody

    @property
    def output(self) -> BaseOutput:
        return self.body.output

    @property
    def invoked_ports(self) -> InvokedPorts:
        return self.body.invoked_ports

    @model_serializer(mode="plain", when_used="json")
    def serialize_model(self) -> Any:
        serialized = super().serialize_model()
        if (
            "body" in serialized
            and isinstance(serialized["body"], dict)
            and "invoked_ports" in serialized["body"]
            and serialized["body"]["invoked_ports"] is None
        ):
            del serialized["body"]["invoked_ports"]
        return serialized


class NodeExecutionFulfilledBody(_BaseNodeExecutionBody, Generic[OutputsType]):
    outputs: OutputsType
    invoked_ports: InvokedPorts = None
    mocked: Optional[bool] = None

    @field_serializer("outputs")
    def serialize_outputs(self, outputs: OutputsType, _info: Any) -> Dict[str, Any]:
        return default_serializer(outputs)

    @field_serializer("invoked_ports")
    def serialize_invoked_ports(self, invoked_ports: InvokedPorts, _info: Any) -> Optional[List[Dict[str, Any]]]:
        if invoked_ports is None:
            return None
        return [default_serializer(port) for port in invoked_ports]


class NodeExecutionFulfilledEvent(_BaseNodeEvent, Generic[OutputsType]):
    name: Literal["node.execution.fulfilled"] = "node.execution.fulfilled"
    body: NodeExecutionFulfilledBody[OutputsType]

    @property
    def outputs(self) -> OutputsType:
        return self.body.outputs

    @property
    def invoked_ports(self) -> InvokedPorts:
        return self.body.invoked_ports

    @property
    def mocked(self) -> Optional[bool]:
        return self.body.mocked

    @model_serializer(mode="plain", when_used="json")
    def serialize_model(self) -> Any:
        serialized = super().serialize_model()
        if (
            "body" in serialized
            and isinstance(serialized["body"], dict)
            and "invoked_ports" in serialized["body"]
            and serialized["body"]["invoked_ports"] is None
        ):
            del serialized["body"]["invoked_ports"]
        return serialized


class NodeExecutionRejectedBody(_BaseNodeExecutionBody):
    error: WorkflowError


class NodeExecutionRejectedEvent(_BaseNodeEvent):
    name: Literal["node.execution.rejected"] = "node.execution.rejected"
    body: NodeExecutionRejectedBody

    @property
    def error(self) -> WorkflowError:
        return self.body.error


class NodeExecutionPausedBody(_BaseNodeExecutionBody):
    pass


class NodeExecutionPausedEvent(_BaseNodeEvent):
    name: Literal["node.execution.paused"] = "node.execution.paused"
    body: NodeExecutionPausedBody


class NodeExecutionResumedBody(_BaseNodeExecutionBody):
    pass


class NodeExecutionResumedEvent(_BaseNodeEvent):
    name: Literal["node.execution.resumed"] = "node.execution.resumed"
    body: NodeExecutionResumedBody


NodeEvent = Union[
    NodeExecutionInitiatedEvent,
    NodeExecutionStreamingEvent,
    NodeExecutionFulfilledEvent,
    NodeExecutionRejectedEvent,
    NodeExecutionPausedEvent,
    NodeExecutionResumedEvent,
]



from typing import TYPE_CHECKING, Iterator, List, Optional, Set, Type, Union

from orderly_set import OrderedSet

from vellum.workflows.edges.edge import Edge
from vellum.workflows.types.generics import NodeType

if TYPE_CHECKING:
    from vellum.workflows.nodes.bases.base import BaseNode
    from vellum.workflows.ports.port import Port

GraphTargetOfSets = Union[
    Set[NodeType],
    Set["Graph"],
    Set["Port"],
    Set[Union[Type["BaseNode"], "Graph", "Port"]],
]

GraphTarget = Union[
    Type["BaseNode"],
    "Port",
    "Graph",
    GraphTargetOfSets,
]


class Graph:
    _entrypoints: Set["Port"]
    _edges: List[Edge]
    _terminals: Set["Port"]

    def __init__(self, entrypoints: Set["Port"], edges: List[Edge], terminals: Set["Port"]):
        self._edges = edges
        self._entrypoints = entrypoints
        self._terminals = terminals

    @staticmethod
    def from_port(port: "Port") -> "Graph":
        ports = {port}
        return Graph(entrypoints=ports, edges=[], terminals=ports)

    @staticmethod
    def from_node(node: Type["BaseNode"]) -> "Graph":
        ports = {port for port in node.Ports}
        return Graph(entrypoints=ports, edges=[], terminals=ports)

    @staticmethod
    def from_set(targets: GraphTargetOfSets) -> "Graph":
        entrypoints = set()
        edges = OrderedSet[Edge]()
        terminals = set()

        for target in targets:
            if isinstance(target, Graph):
                entrypoints.update(target._entrypoints)
                edges.update(target._edges)
                terminals.update(target._terminals)
            elif hasattr(target, "Ports"):
                entrypoints.update({port for port in target.Ports})
                terminals.update({port for port in target.Ports})
            else:
                # target is a Port
                entrypoints.update({target})
                terminals.update({target})

        return Graph(entrypoints=entrypoints, edges=list(edges), terminals=terminals)

    @staticmethod
    def from_edge(edge: Edge) -> "Graph":
        return Graph(entrypoints={edge.from_port}, edges=[edge], terminals={port for port in edge.to_node.Ports})

    def __rshift__(self, other: GraphTarget) -> "Graph":
        if not self._edges and not self._entrypoints:
            raise ValueError("Graph instance can only create new edges from nodes within existing edges")

        if isinstance(other, set):
            new_terminals = set()
            for elem in other:
                for final_output_node in self._terminals:
                    if isinstance(elem, Graph):
                        midgraph = final_output_node >> set(elem.entrypoints)
                        self._extend_edges(midgraph.edges)
                        self._extend_edges(elem.edges)
                        for other_terminal in elem._terminals:
                            new_terminals.add(other_terminal)
                    elif hasattr(elem, "Ports"):
                        midgraph = final_output_node >> elem
                        self._extend_edges(midgraph.edges)
                        for other_terminal in elem.Ports:
                            new_terminals.add(other_terminal)
                    else:
                        # elem is a Port
                        midgraph = final_output_node >> elem
                        self._extend_edges(midgraph.edges)
                        new_terminals.add(elem)
            self._terminals = new_terminals
            return self

        if isinstance(other, Graph):
            for final_output_node in self._terminals:
                midgraph = final_output_node >> set(other.entrypoints)
                self._extend_edges(midgraph.edges)
                self._extend_edges(other.edges)
            self._terminals = other._terminals
            return self

        if hasattr(other, "Ports"):
            for final_output_node in self._terminals:
                subgraph = final_output_node >> other
                self._extend_edges(subgraph.edges)
            self._terminals = {port for port in other.Ports}
            return self

        # other is a Port
        for final_output_node in self._terminals:
            subgraph = final_output_node >> other
            self._extend_edges(subgraph.edges)
        self._terminals = {other}
        return self

    def __rrshift__(cls, other_cls: GraphTarget) -> "Graph":
        if not isinstance(other_cls, set):
            other_cls = {other_cls}

        return Graph.from_set(other_cls) >> cls

    @property
    def entrypoints(self) -> Iterator[Type["BaseNode"]]:
        return iter(e.node_class for e in self._entrypoints)

    @property
    def edges(self) -> Iterator[Edge]:
        return iter(self._edges)

    @property
    def nodes(self) -> Iterator[Type["BaseNode"]]:
        nodes = set()
        if not self._edges:
            for node in self.entrypoints:
                if node not in nodes:
                    nodes.add(node)
                    yield node
            return

        for edge in self._edges:
            if edge.from_port.node_class not in nodes:
                nodes.add(edge.from_port.node_class)
                yield edge.from_port.node_class
            if edge.to_node not in nodes:
                nodes.add(edge.to_node)
                yield edge.to_node

    def _extend_edges(self, edges: Iterator[Edge]) -> None:
        for edge in edges:
            if edge not in self._edges:
                self._edges.append(edge)

    def __str__(self) -> str:
        """
        Return a visual ASCII representation of the graph showing the flow structure.
        """
        if not self._edges and not self._entrypoints:
            return "Graph(empty)"

        if not self._edges:
            if len(self._entrypoints) == 1:
                port = next(iter(self._entrypoints))
                return f"Graph: {self._get_port_name(port)}"
            else:
                port_names = [self._get_port_name(port) for port in self._entrypoints]
                return f"Graph: [{', '.join(port_names)}]"

        return self._build_flow_diagram()

    def _build_flow_diagram(self) -> str:
        """Build a connected flow diagram showing the graph structure."""
        lines = ["Graph:"]

        adjacency: dict[str, list[str]] = {}
        all_nodes = set()

        for edge in self._edges:
            source_node = edge.from_port.node_class.__name__
            target_node = edge.to_node.__name__

            all_nodes.add(source_node)
            all_nodes.add(target_node)

            if source_node not in adjacency:
                adjacency[source_node] = []
            adjacency[source_node].append(target_node)

        target_nodes = set()
        for edges in adjacency.values():
            target_nodes.update(edges)

        root_nodes = []
        for port in self._entrypoints:
            node_name = port.node_class.__name__
            if node_name not in target_nodes:
                root_nodes.append(node_name)

        if not root_nodes and self._entrypoints:
            root_nodes = [next(iter(self._entrypoints)).node_class.__name__]

        visited = set()
        currently_visiting = set()

        def render_node(node: str, prefix: str = "  ", is_last: bool = True, path: Optional[List[str]] = None) -> None:
            if path is None:
                path = []

            if node in currently_visiting:
                lines.append(f"{prefix}{'└─' if is_last else '├─'} {node} ⟲ (loops back)")
                return

            if node in visited:
                lines.append(f"{prefix}{'└─' if is_last else '├─'} {node} → (see above)")
                return

            visited.add(node)
            currently_visiting.add(node)

            lines.append(f"{prefix}{'└─' if is_last else '├─'} {node}")

            if node in adjacency:
                children = adjacency[node]
                for i, child in enumerate(children):
                    child_is_last = i == len(children) - 1
                    next_prefix = prefix + ("   " if is_last else "│  ")
                    render_node(child, next_prefix, child_is_last, path + [node])

            currently_visiting.remove(node)

        for i, root in enumerate(root_nodes):
            is_last_root = i == len(root_nodes) - 1
            render_node(root, "  ", is_last_root)

        return "\n".join(lines)

    def _get_port_name(self, port: "Port") -> str:
        """Get a readable name for a port."""
        try:
            if hasattr(port, "node_class") and hasattr(port.node_class, "__name__"):
                node_name = port.node_class.__name__
                port_name = getattr(port, "name", "unknown")
                return f"{node_name}.{port_name}"
            else:
                return str(port)
        except Exception:
            return f"<Port:{getattr(port, 'name', 'unknown')}>"

    def _get_node_name(self, node: Type["BaseNode"]) -> str:
        """Get a readable name for a node."""
        try:
            return getattr(node, "__name__", str(node))
        except Exception:
            return "<Node:unknown>"



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.paginated_folder_entity_list import PaginatedFolderEntityList
from .raw_client import AsyncRawFolderEntitiesClient, RawFolderEntitiesClient
from .types.folder_entities_list_request_entity_status import FolderEntitiesListRequestEntityStatus

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class FolderEntitiesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawFolderEntitiesClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawFolderEntitiesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawFolderEntitiesClient
        """
        return self._raw_client

    def list(
        self,
        *,
        parent_folder_id: str,
        entity_status: typing.Optional[FolderEntitiesListRequestEntityStatus] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedFolderEntityList:
        """
        List all folder entities within a specified folder.

        Parameters
        ----------
        parent_folder_id : str
            Filter down to only those entities whose parent folder has the specified ID.

            To filter by an entity's parent folder, provide the ID of the parent folder. To filter by the root directory, provide
            a string representing the entity type of the root directory. Supported root directories include:
            - PROMPT_SANDBOX
            - WORKFLOW_SANDBOX
            - DOCUMENT_INDEX
            - TEST_SUITE

        entity_status : typing.Optional[FolderEntitiesListRequestEntityStatus]
            Filter down to only those objects whose entities have a status matching the status specified.

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedFolderEntityList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.folder_entities.list(
            parent_folder_id="parent_folder_id",
        )
        """
        _response = self._raw_client.list(
            parent_folder_id=parent_folder_id,
            entity_status=entity_status,
            limit=limit,
            offset=offset,
            ordering=ordering,
            request_options=request_options,
        )
        return _response.data

    def add_entity_to_folder(
        self, folder_id: str, *, entity_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Add an entity to a specific folder or root directory.

        Adding an entity to a folder will remove it from any other folders it might have been a member of.

        Parameters
        ----------
        folder_id : str
            The ID of the folder to which the entity should be added. This can be a UUID of a folder, or the name of a root
            directory. Supported root directories include:
            - PROMPT_SANDBOX
            - WORKFLOW_SANDBOX
            - DOCUMENT_INDEX
            - TEST_SUITE

        entity_id : str
            The ID of the entity you would like to move.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.folder_entities.add_entity_to_folder(
            folder_id="folder_id",
            entity_id="entity_id",
        )
        """
        _response = self._raw_client.add_entity_to_folder(
            folder_id, entity_id=entity_id, request_options=request_options
        )
        return _response.data


class AsyncFolderEntitiesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawFolderEntitiesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawFolderEntitiesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawFolderEntitiesClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        parent_folder_id: str,
        entity_status: typing.Optional[FolderEntitiesListRequestEntityStatus] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedFolderEntityList:
        """
        List all folder entities within a specified folder.

        Parameters
        ----------
        parent_folder_id : str
            Filter down to only those entities whose parent folder has the specified ID.

            To filter by an entity's parent folder, provide the ID of the parent folder. To filter by the root directory, provide
            a string representing the entity type of the root directory. Supported root directories include:
            - PROMPT_SANDBOX
            - WORKFLOW_SANDBOX
            - DOCUMENT_INDEX
            - TEST_SUITE

        entity_status : typing.Optional[FolderEntitiesListRequestEntityStatus]
            Filter down to only those objects whose entities have a status matching the status specified.

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedFolderEntityList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.folder_entities.list(
                parent_folder_id="parent_folder_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.list(
            parent_folder_id=parent_folder_id,
            entity_status=entity_status,
            limit=limit,
            offset=offset,
            ordering=ordering,
            request_options=request_options,
        )
        return _response.data

    async def add_entity_to_folder(
        self, folder_id: str, *, entity_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Add an entity to a specific folder or root directory.

        Adding an entity to a folder will remove it from any other folders it might have been a member of.

        Parameters
        ----------
        folder_id : str
            The ID of the folder to which the entity should be added. This can be a UUID of a folder, or the name of a root
            directory. Supported root directories include:
            - PROMPT_SANDBOX
            - WORKFLOW_SANDBOX
            - DOCUMENT_INDEX
            - TEST_SUITE

        entity_id : str
            The ID of the entity you would like to move.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.folder_entities.add_entity_to_folder(
                folder_id="folder_id",
                entity_id="entity_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.add_entity_to_folder(
            folder_id, entity_id=entity_id, request_options=request_options
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .scenario_input import ScenarioInput


class SandboxScenario(UniversalBaseModel):
    """
    Sandbox Scenario
    """

    label: typing.Optional[str] = None
    inputs: typing.List[ScenarioInput] = pydantic.Field()
    """
    The inputs for the scenario
    """

    id: str = pydantic.Field()
    """
    The id of the scenario
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from vellum import ChatMessagePromptBlock, PlainTextPromptBlock, RichTextPromptBlock


def TextSystemMessage(content: str) -> ChatMessagePromptBlock:
    """
    Create a text system message that autocasts to ChatMessagePromptBlock.

    Args:
        content: The text content for the system message

    Returns:
        ChatMessagePromptBlock configured as a system message
    """
    return ChatMessagePromptBlock(
        chat_role="SYSTEM", blocks=[RichTextPromptBlock(blocks=[PlainTextPromptBlock(text=content)])]
    )


def TextUserMessage(content: str) -> ChatMessagePromptBlock:
    """
    Create a text user message that autocasts to ChatMessagePromptBlock.

    Args:
        content: The text content for the user message

    Returns:
        ChatMessagePromptBlock configured as a user message
    """
    return ChatMessagePromptBlock(
        chat_role="USER", blocks=[RichTextPromptBlock(blocks=[PlainTextPromptBlock(text=content)])]
    )



from typing import List

from deepdiff import DeepDiff

from vellum import ChatMessagePromptBlock, FunctionDefinition, JinjaPromptBlock
from vellum.client.types.chat_message import ChatMessage
from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import InlinePromptNode
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state import BaseState
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_inline_prompt_node_with_functions.workflow import BasicInlinePromptWithFunctionsWorkflow


def test_serialize_workflow():
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicInlinePromptWithFunctionsWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "ceb5cc94-48ee-4968-b37a-421623a8f1ef",
                "key": "noun",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            }
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [{"id": "15a0ab89-8ed4-43b9-afa2-3c0b29d4dc3e", "key": "results", "type": "JSON"}],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 2
    assert len(workflow_raw_data["nodes"]) == 3

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "382842a3-0490-4dee-b87b-eef86766f07c",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "8294baa6-8bf4-4b54-a56b-407b64851b77"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    prompt_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "8450dd06-975a-41a4-a564-808ee8808fe6",
            "type": "PROMPT",
            "inputs": [
                {
                    "id": "f7fca55e-93e9-4009-9227-acf839c7990d",
                    "key": "noun",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "ceb5cc94-48ee-4968-b37a-421623a8f1ef"},
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "data": {
                "label": "Example Base Inline Prompt Node With Functions",
                "output_id": "ead0ccb5-092f-4d9b-a9ec-5eb83d498188",
                "error_output_id": None,
                "array_output_id": "628df199-a049-40b9-a29b-a378edd759bb",
                "source_handle_id": "d4a097ab-e22d-42f1-b6bc-2ed96856377a",
                "target_handle_id": "c2dccecb-8a41-40a8-95af-325d3ab8bfe5",
                "variant": "INLINE",
                "exec_config": {
                    "parameters": {
                        "stop": [],
                        "temperature": 0.0,
                        "max_tokens": 4096,
                        "top_p": 1.0,
                        "top_k": 0,
                        "frequency_penalty": 0.0,
                        "presence_penalty": 0.0,
                        "logit_bias": None,
                        "custom_parameters": None,
                    },
                    "input_variables": [
                        {"id": "f7fca55e-93e9-4009-9227-acf839c7990d", "key": "noun", "type": "STRING"}
                    ],
                    "prompt_template_block_data": {
                        "version": 1,
                        "blocks": [
                            {
                                "block_type": "CHAT_MESSAGE",
                                "properties": {
                                    "chat_role": "SYSTEM",
                                    "chat_source": None,
                                    "chat_message_unterminated": False,
                                    "blocks": [
                                        {
                                            "block_type": "JINJA",
                                            "properties": {
                                                "template": "What's your favorite {{noun}}?",
                                                "template_type": "STRING",
                                            },
                                            "id": "467fe2b1-312b-40db-8869-9c6ada7c7077",
                                            "cache_config": None,
                                            "state": "ENABLED",
                                        }
                                    ],
                                },
                                "id": "1d1e117d-19dc-4282-b1e3-9534014fb6e5",
                                "cache_config": None,
                                "state": "ENABLED",
                            },
                            {
                                "id": "9b34f084-449d-423f-8691-37518b1ee9ca",
                                "block_type": "FUNCTION_DEFINITION",
                                "properties": {
                                    "function_name": "favorite_noun",
                                    "function_description": "Returns the favorite noun of the user",
                                    "function_parameters": {},
                                    "function_forced": None,
                                    "function_strict": None,
                                },
                            },
                        ],
                    },
                },
                "ml_model_name": "gpt-4o",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "InlinePromptNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "inline_prompt_node", "node"],
            },
            "definition": {
                "name": "ExampleBaseInlinePromptNodeWithFunctions",
                "module": ["tests", "workflows", "basic_inline_prompt_node_with_functions", "workflow"],
            },
            "trigger": {
                "id": "c2dccecb-8a41-40a8-95af-325d3ab8bfe5",
                "merge_behavior": "AWAIT_ANY",
            },
            "outputs": [
                {"id": "9557bd86-702d-4b45-b8c1-c3980bffe28f", "name": "json", "type": "JSON", "value": None},
                {"id": "ead0ccb5-092f-4d9b-a9ec-5eb83d498188", "name": "text", "type": "STRING", "value": None},
                {"id": "628df199-a049-40b9-a29b-a378edd759bb", "name": "results", "type": "ARRAY", "value": None},
            ],
            "ports": [{"id": "d4a097ab-e22d-42f1-b6bc-2ed96856377a", "name": "default", "type": "DEFAULT"}],
            "attributes": [
                {
                    "id": "6cd5395c-6e46-4bc9-b98c-8f8924554555",
                    "name": "ml_model",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "gpt-4o"}},
                },
                {
                    "id": "25f935f3-363f-4ead-a5a0-db234ca67e1e",
                    "name": "blocks",
                    "value": {
                        "type": "CONSTANT_VALUE",
                        "value": {
                            "type": "JSON",
                            "value": [
                                {
                                    "block_type": "CHAT_MESSAGE",
                                    "state": None,
                                    "cache_config": None,
                                    "chat_role": "SYSTEM",
                                    "chat_source": None,
                                    "chat_message_unterminated": None,
                                    "blocks": [
                                        {
                                            "block_type": "JINJA",
                                            "state": None,
                                            "cache_config": None,
                                            "template": "What's your favorite {{noun}}?",
                                        }
                                    ],
                                }
                            ],
                        },
                    },
                },
                {
                    "id": "ffabe7d2-8ab6-4201-9d41-c4d7be1386e1",
                    "name": "prompt_inputs",
                    "value": {
                        "type": "DICTIONARY_REFERENCE",
                        "entries": [
                            {
                                "id": "0bfa70a2-164f-460c-9e9a-4d62221eadf4",
                                "key": "noun",
                                "value": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": "ceb5cc94-48ee-4968-b37a-421623a8f1ef",
                                },
                            }
                        ],
                    },
                },
                {
                    "id": "8107682b-2ca0-4967-88f9-284455936575",
                    "name": "functions",
                    "value": {
                        "type": "CONSTANT_VALUE",
                        "value": {
                            "type": "JSON",
                            "value": [
                                {
                                    "state": None,
                                    "cache_config": None,
                                    "name": "favorite_noun",
                                    "description": "Returns the favorite noun of the user",
                                    "parameters": {},
                                    "forced": None,
                                    "strict": None,
                                }
                            ],
                        },
                    },
                },
                {
                    "id": "2b98319f-f43d-42d9-a8b0-b148d5de0a2c",
                    "name": "parameters",
                    "value": {
                        "type": "CONSTANT_VALUE",
                        "value": {
                            "type": "JSON",
                            "value": {
                                "stop": [],
                                "temperature": 0.0,
                                "max_tokens": 4096,
                                "top_p": 1.0,
                                "top_k": 0,
                                "frequency_penalty": 0.0,
                                "presence_penalty": 0.0,
                                "logit_bias": None,
                                "custom_parameters": None,
                            },
                        },
                    },
                },
            ],
        },
        prompt_node,
        ignore_order=True,
    )

    final_output_node = workflow_raw_data["nodes"][2]
    assert not DeepDiff(
        {
            "id": "42318326-3ae8-417f-9609-f6d8ae47eafb",
            "type": "TERMINAL",
            "data": {
                "label": "Final Output",
                "name": "results",
                "target_handle_id": "46c99277-2b4b-477d-851c-ea497aef6b16",
                "output_id": "15a0ab89-8ed4-43b9-afa2-3c0b29d4dc3e",
                "output_type": "JSON",
                "node_input_id": "d7c89dce-765b-494d-a256-aba4bcf87b42",
            },
            "inputs": [
                {
                    "id": "d7c89dce-765b-494d-a256-aba4bcf87b42",
                    "key": "node_input",
                    "value": {
                        "rules": [
                            {
                                "type": "NODE_OUTPUT",
                                "data": {
                                    "node_id": "8450dd06-975a-41a4-a564-808ee8808fe6",
                                    "output_id": "628df199-a049-40b9-a29b-a378edd759bb",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "display_data": {"position": {"x": 400.0, "y": -50.0}},
            "base": {
                "name": "FinalOutputNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
            },
            "definition": None,
        },
        final_output_node,
        ignore_order=True,
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "924f693f-3f4c-466a-8cde-648ba3baf9fd",
                "source_node_id": "382842a3-0490-4dee-b87b-eef86766f07c",
                "source_handle_id": "8294baa6-8bf4-4b54-a56b-407b64851b77",
                "target_node_id": "8450dd06-975a-41a4-a564-808ee8808fe6",
                "target_handle_id": "c2dccecb-8a41-40a8-95af-325d3ab8bfe5",
                "type": "DEFAULT",
            },
            {
                "id": "05ca58fb-e02d-48d4-9207-2dad0833a25b",
                "source_node_id": "8450dd06-975a-41a4-a564-808ee8808fe6",
                "source_handle_id": "d4a097ab-e22d-42f1-b6bc-2ed96856377a",
                "target_node_id": "42318326-3ae8-417f-9609-f6d8ae47eafb",
                "target_handle_id": "46c99277-2b4b-477d-851c-ea497aef6b16",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicInlinePromptWithFunctionsWorkflow",
        "module": ["tests", "workflows", "basic_inline_prompt_node_with_functions", "workflow"],
    }


def test_serialize_workflow_with_descriptor_functions():
    """Test that serialization handles BaseDescriptor instances in functions list."""

    class TestInputs(BaseInputs):
        noun: str

    class MockMCPClientNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            tools: list[FunctionDefinition]

    class TestInlinePromptNodeWithDescriptorFunctions(InlinePromptNode):
        ml_model = "gpt-4o"
        blocks = [
            ChatMessagePromptBlock(
                chat_role="SYSTEM",
                blocks=[JinjaPromptBlock(template="Test {{noun}}")],
            ),
        ]
        prompt_inputs = {"noun": TestInputs.noun}
        functions = MockMCPClientNode.Outputs.tools  # type: ignore

    class TestWorkflow(BaseWorkflow[TestInputs, BaseState]):
        graph = MockMCPClientNode >> TestInlinePromptNodeWithDescriptorFunctions

    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized: dict = workflow_display.serialize()

    prompt_nodes = [node for node in serialized["workflow_raw_data"]["nodes"] if node["type"] == "PROMPT"]
    assert len(prompt_nodes) == 1

    prompt_node = prompt_nodes[0]
    assert isinstance(prompt_node, dict)
    blocks = prompt_node["data"]["exec_config"]["prompt_template_block_data"]["blocks"]
    assert isinstance(blocks, list)

    function_blocks = [
        block for block in blocks if isinstance(block, dict) and block.get("block_type") == "FUNCTION_DEFINITION"
    ]
    assert len(function_blocks) == 0  # We don't serialize the legacy function blocks when dynamic

    assert "attributes" in prompt_node
    assert isinstance(prompt_node["attributes"], list)
    functions_attr = next(
        (attr for attr in prompt_node["attributes"] if isinstance(attr, dict) and attr["name"] == "functions"), None
    )
    assert isinstance(functions_attr, dict), "functions attribute should be present in serialized attributes"

    assert functions_attr["value"] == {
        "node_id": "cb1186e0-8ff1-4145-823e-96b3fc05a39a",
        "node_output_id": "470fadb9-b8b5-477e-a502-5209d398bcf9",
        "type": "NODE_OUTPUT",
    }


def test_serialize_workflow_with_descriptor_blocks():
    """Test that serialization handles BaseDescriptor instances in blocks list."""

    class TestInputs(BaseInputs):
        noun: str

    class UpstreamNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            results: list

        def run(self) -> Outputs:
            return self.Outputs(results=["test"])

    class TestInlinePromptNodeWithDescriptorBlocks(InlinePromptNode):
        ml_model = "gpt-4o"
        blocks = [UpstreamNode.Outputs.results[0]]  # type: ignore
        prompt_inputs = {"noun": TestInputs.noun}

    class TestWorkflow(BaseWorkflow[TestInputs, BaseState]):
        graph = UpstreamNode >> TestInlinePromptNodeWithDescriptorBlocks

    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized: dict = workflow_display.serialize()

    prompt_nodes = [node for node in serialized["workflow_raw_data"]["nodes"] if node["type"] == "PROMPT"]
    prompt_node = prompt_nodes[0]

    blocks = prompt_node["data"]["exec_config"]["prompt_template_block_data"]["blocks"]
    descriptor_blocks = [block for block in blocks if not isinstance(block, dict) or not block.get("block_type")]
    assert len(descriptor_blocks) == 0, "BaseDescriptor blocks should not appear in serialized blocks"

    blocks_attr = next((attr for attr in prompt_node["attributes"] if attr["name"] == "blocks"), None)
    assert blocks_attr is not None, "blocks attribute should be present when blocks contain BaseDescriptor"
    assert blocks_attr["value"]["type"] == "ARRAY_REFERENCE", "blocks attribute should be serialized as ARRAY_REFERENCE"
    assert blocks_attr["value"]["items"] == [
        {
            "type": "BINARY_EXPRESSION",
            "lhs": {
                "type": "NODE_OUTPUT",
                "node_id": str(UpstreamNode.__id__),
                "node_output_id": str(UpstreamNode.__output_ids__["results"]),
            },
            "operator": "accessField",
            "rhs": {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "NUMBER",
                    "value": 0.0,
                },
            },
        }
    ]


def test_serialize_workflow_with_nested_descriptor_blocks():
    """Test that serialization handles BaseDescriptor instances nested in ChatMessageBlock.blocks."""

    class TestInputs(BaseInputs):
        noun: str

    class UpstreamNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            results: list

        def run(self) -> Outputs:
            return self.Outputs(results=["test"])

    chat_block = ChatMessagePromptBlock(chat_role="SYSTEM", blocks=[JinjaPromptBlock(template="Hello")])

    class TestInlinePromptNodeWithNestedDescriptorBlocks(InlinePromptNode):
        ml_model = "gpt-4o"
        blocks = [chat_block]
        prompt_inputs = {"noun": TestInputs.noun}

    object.__setattr__(chat_block, "blocks", [UpstreamNode.Outputs.results[0]])

    class TestWorkflow(BaseWorkflow[TestInputs, BaseState]):
        graph = UpstreamNode >> TestInlinePromptNodeWithNestedDescriptorBlocks

    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized: dict = workflow_display.serialize()

    prompt_nodes = [node for node in serialized["workflow_raw_data"]["nodes"] if node["type"] == "PROMPT"]
    prompt_node = prompt_nodes[0]

    blocks = prompt_node["data"]["exec_config"]["prompt_template_block_data"]["blocks"]
    descriptor_blocks = [block for block in blocks if not isinstance(block, dict) or not block.get("block_type")]
    assert len(descriptor_blocks) == 0, "BaseDescriptor blocks should not appear in serialized blocks"

    blocks_attr = next((attr for attr in prompt_node["attributes"] if attr["name"] == "blocks"), None)
    assert blocks_attr is not None, "blocks attribute should be present when blocks contain nested BaseDescriptor"
    assert blocks_attr["value"]["type"] == "ARRAY_REFERENCE", "blocks attribute should be serialized as ARRAY_REFERENCE"
    assert blocks_attr["value"]["items"] == [
        {
            "entries": [
                {
                    "id": "24a203be-3cba-4b20-bc84-9993a476c120",
                    "key": "block_type",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "CHAT_MESSAGE"}},
                },
                {
                    "id": "c06269e6-f74c-4860-8fa5-22dcbdc89399",
                    "key": "state",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
                },
                {
                    "id": "dd9c0d43-b931-4dc8-8b3a-a7507ddff0c1",
                    "key": "cache_config",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
                },
                {
                    "id": "bef22f2b-0b6e-4910-88cc-6df736d2e20e",
                    "key": "chat_role",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "SYSTEM"}},
                },
                {
                    "id": "c0beec30-f85e-4a78-a3fb-baee54a692f8",
                    "key": "chat_source",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
                },
                {
                    "id": "f601f4f2-62fe-4697-9fe0-99ca8aa64500",
                    "key": "chat_message_unterminated",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
                },
                {
                    "id": "ad550008-64e3-44a3-a32a-84ec226db31c",
                    "key": "blocks",
                    "value": {
                        "items": [
                            {
                                "lhs": {
                                    "node_id": "9fe5d3a3-7d26-4692-aa2d-e67c673b0c2b",
                                    "node_output_id": "92f9a1b7-d33b-4f00-b4c2-e6f58150e166",
                                    "type": "NODE_OUTPUT",
                                },
                                "operator": "accessField",
                                "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 0.0}},
                                "type": "BINARY_EXPRESSION",
                            }
                        ],
                        "type": "ARRAY_REFERENCE",
                    },
                },
            ],
            "type": "DICTIONARY_REFERENCE",
        }
    ]


def test_inline_prompt_node__coalesce_expression_serialization():
    """
    Tests that prompt nodes can serialize coalesce expressions like State.chat_history.coalesce([]).
    """

    # GIVEN a custom state with chat_history
    class MyState(BaseState):
        chat_history: List[ChatMessage] = []

    # AND a prompt node that uses a coalesce expression as input
    class MyNode(InlinePromptNode[MyState]):
        ml_model = "gpt-4o"
        blocks = []
        prompt_inputs = {
            "chat_history": MyState.chat_history.coalesce([]),
        }

    class TestWorkflow(BaseWorkflow[BaseInputs, MyState]):
        graph = MyNode

    # WHEN the node is serialized
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized: dict = workflow_display.serialize()

    # THEN the prompt is serialized with the correct inputs
    prompt_nodes = [node for node in serialized["workflow_raw_data"]["nodes"] if node["type"] == "PROMPT"]
    prompt_node = prompt_nodes[0]

    prompt_inputs_attr = next((attr for attr in prompt_node["attributes"] if attr["name"] == "prompt_inputs"), None)
    assert prompt_inputs_attr
    assert prompt_inputs_attr["value"]["type"] == "DICTIONARY_REFERENCE"
    chat_history_entry = prompt_inputs_attr["value"]["entries"][0]

    assert chat_history_entry["key"] == "chat_history"
    assert chat_history_entry["value"]["type"] == "BINARY_EXPRESSION"
    assert chat_history_entry["value"]["operator"] == "coalesce"
    assert chat_history_entry["value"]["lhs"] == {
        "type": "WORKFLOW_STATE",
        "state_variable_id": "6012a4f7-a8ff-464d-bd62-7c41fde06fa4",
    }
    assert chat_history_entry["value"]["rhs"] == {
        "type": "CONSTANT_VALUE",
        "value": {
            "type": "JSON",
            "value": [],
        },
    }



from vellum import TestSuiteRunMetricNumberOutput
from vellum.client.types.chat_history_input import ChatHistoryInput
from vellum.client.types.chat_message import ChatMessage
from vellum.client.types.json_input import JsonInput
from vellum.client.types.metric_definition_execution import MetricDefinitionExecution
from vellum.client.types.number_input import NumberInput
from vellum.client.types.string_input import StringInput
from vellum.workflows.nodes.displayable.guardrail_node.node import GuardrailNode


def test_guardrail_node__inputs(vellum_client):
    """Test that GuardrailNode correctly handles inputs."""

    # GIVEN a Guardrail Node with inputs
    class MyGuard(GuardrailNode):
        metric_definition = "example_metric_definition"
        metric_inputs = {
            "a_string": "hello",
            "a_chat_history": [ChatMessage(role="USER", text="Hello, how are you?")],
            "a_dict": {"foo": "bar"},
            "a_int": 42,
            "a_float": 3.14,
        }

    vellum_client.metric_definitions.execute_metric_definition.return_value = MetricDefinitionExecution(
        outputs=[
            TestSuiteRunMetricNumberOutput(
                name="score",
                value=1.0,
            ),
        ],
    )

    # WHEN the node is run
    MyGuard().run()

    # THEN the metric_definitions.execute_metric_definition method should be called with the correct inputs
    mock_api = vellum_client.metric_definitions.execute_metric_definition
    assert mock_api.call_count == 1

    assert mock_api.call_args.kwargs["inputs"] == [
        StringInput(name="a_string", type="STRING", value="hello"),
        ChatHistoryInput(
            name="a_chat_history", type="CHAT_HISTORY", value=[ChatMessage(role="USER", text="Hello, how are you?")]
        ),
        JsonInput(name="a_dict", type="JSON", value={"foo": "bar"}),
        NumberInput(name="a_int", type="NUMBER", value=42.0),
        NumberInput(name="a_float", type="NUMBER", value=3.14),
    ]
    assert len(mock_api.call_args.kwargs["inputs"]) == 5



class NodeValidationError(Exception):
    """Exception raised when a node fails validation during workflow display serialization."""

    def __init__(self, message: str, node_class_name: str):
        self.message = message
        self.node_class_name = node_class_name
        super().__init__(f"Node validation error in {node_class_name}: {message}")



import json
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, Union, cast

from pydash import snake_case

from vellum import ChatMessage, PromptBlock
from vellum.client.types.array_chat_message_content import ArrayChatMessageContent
from vellum.client.types.array_chat_message_content_item import ArrayChatMessageContentItem
from vellum.client.types.function_call_chat_message_content import FunctionCallChatMessageContent
from vellum.client.types.function_call_chat_message_content_value import FunctionCallChatMessageContentValue
from vellum.client.types.function_definition import FunctionDefinition
from vellum.client.types.prompt_output import PromptOutput
from vellum.client.types.prompt_parameters import PromptParameters
from vellum.client.types.string_chat_message_content import StringChatMessageContent
from vellum.client.types.variable_prompt_block import VariablePromptBlock
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.integrations.composio_service import ComposioService
from vellum.workflows.integrations.mcp_service import MCPService
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.core.inline_subworkflow_node.node import InlineSubworkflowNode
from vellum.workflows.nodes.displayable.inline_prompt_node.node import InlinePromptNode
from vellum.workflows.nodes.displayable.subworkflow_deployment_node.node import SubworkflowDeploymentNode
from vellum.workflows.nodes.displayable.tool_calling_node.state import ToolCallingState
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.ports.port import Port
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.state import BaseState
from vellum.workflows.state.encoder import DefaultStateEncoder
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior, Tool, ToolBase
from vellum.workflows.types.definition import ComposioToolDefinition, DeploymentDefinition, MCPServer, MCPToolDefinition
from vellum.workflows.types.generics import is_workflow_class

CHAT_HISTORY_VARIABLE = "chat_history"


logger = logging.getLogger(__name__)


class FunctionCallNodeMixin:
    """Mixin providing common functionality for nodes that handle function calls."""

    function_call_output: List[PromptOutput]

    def _extract_function_arguments(self) -> dict:
        """Extract arguments from function call output."""
        current_index = getattr(self, "state").current_prompt_output_index
        if self.function_call_output and len(self.function_call_output) > current_index:
            function_call = self.function_call_output[current_index]
            if function_call.type == "FUNCTION_CALL" and function_call.value is not None:
                return function_call.value.arguments or {}
        return {}

    def _add_function_result_to_chat_history(self, result: Any, state: ToolCallingState) -> None:
        """Add function execution result to chat history."""
        state.chat_history.append(
            ChatMessage(
                role="FUNCTION",
                content=StringChatMessageContent(value=json.dumps(result, cls=DefaultStateEncoder)),
            )
        )
        with state.__quiet__():
            state.current_function_calls_processed += 1
            state.current_prompt_output_index += 1


class ToolPromptNode(InlinePromptNode[ToolCallingState]):
    max_prompt_iterations: Optional[int] = 5

    class Trigger(InlinePromptNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ATTRIBUTES

    def run(self) -> Iterator[BaseOutput]:
        if self.max_prompt_iterations is not None and self.state.prompt_iterations >= self.max_prompt_iterations:
            max_iterations_message = f"Maximum number of prompt iterations `{self.max_prompt_iterations}` reached."
            raise NodeException(message=max_iterations_message, code=WorkflowErrorCode.NODE_EXECUTION)

        generator = super().run()
        with self.state.__quiet__():
            self.state.current_prompt_output_index = 0
            self.state.current_function_calls_processed = 0
            self.state.prompt_iterations += 1
        for output in generator:
            if output.name == InlinePromptNode.Outputs.results.name and output.value:
                prompt_outputs = cast(List[PromptOutput], output.value)
                chat_contents: List[ArrayChatMessageContentItem] = []
                for prompt_output in prompt_outputs:
                    if prompt_output.type == "STRING":
                        chat_contents.append(StringChatMessageContent(value=prompt_output.value))
                    elif prompt_output.type == "FUNCTION_CALL" and prompt_output.value:
                        raw_function_call = prompt_output.value.model_dump()
                        if "state" in raw_function_call:
                            del raw_function_call["state"]
                        chat_contents.append(
                            FunctionCallChatMessageContent(
                                value=FunctionCallChatMessageContentValue.model_validate(raw_function_call)
                            )
                        )

                if len(chat_contents) == 1:
                    if chat_contents[0].type == "STRING":
                        self.state.chat_history.append(ChatMessage(role="ASSISTANT", text=chat_contents[0].value))
                    else:
                        self.state.chat_history.append(ChatMessage(role="ASSISTANT", content=chat_contents[0]))
                else:
                    self.state.chat_history.append(
                        ChatMessage(role="ASSISTANT", content=ArrayChatMessageContent(value=chat_contents))
                    )

            yield output


class RouterNode(BaseNode[ToolCallingState]):
    """Router node that handles routing to function nodes based on outputs."""

    pass


class DynamicSubworkflowDeploymentNode(SubworkflowDeploymentNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a deployment definition with function call output."""

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        # Mypy doesn't like instance assignments of class attributes. It's safe in our case tho bc it's what
        # we do in the `__init__` method. Long term, instead of the function_call_output attribute above, we
        # want to do:
        # ```python
        # subworkflow_inputs = tool_prompt_node.Outputs.results[0]['value']['arguments'].if_(
        #     tool_prompt_node.Outputs.results[0]['type'].equals('FUNCTION_CALL'),
        #     {},
        # )
        # ```
        self.subworkflow_inputs = arguments  # type:ignore[misc]

        # Call the parent run method to execute the subworkflow
        outputs = {}
        for output in super().run():
            if output.is_fulfilled:
                outputs[output.name] = output.value
            yield output

        # Add the result to the chat history
        self._add_function_result_to_chat_history(outputs, self.state)


class DynamicInlineSubworkflowNode(
    InlineSubworkflowNode[ToolCallingState, BaseInputs, BaseState], FunctionCallNodeMixin
):
    """Node that executes an inline subworkflow with function call output."""

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        self.subworkflow_inputs = arguments  # type: ignore[misc]

        # Call the parent run method to execute the subworkflow with proper streaming
        outputs = {}

        for output in super().run():
            if output.is_fulfilled:
                outputs[output.name] = output.value
            yield output

        # Add the result to the chat history
        self._add_function_result_to_chat_history(outputs, self.state)


class FunctionNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a regular Python function with function call output."""

    function_definition: Callable[..., Any]

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        try:
            result = self.function_definition(**arguments)
        except Exception as e:
            function_name = self.function_definition.__name__
            raise NodeException(
                message=f"Error executing function '{function_name}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            )

        # Add the result to the chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield from []


class ComposioNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a Composio tool with function call output."""

    composio_tool: ComposioToolDefinition

    def run(self) -> Iterator[BaseOutput]:
        # Extract arguments from function call
        arguments = self._extract_function_arguments()

        try:
            # Execute using ComposioService
            composio_service = ComposioService()
            if self.composio_tool.user_id is not None:
                result = composio_service.execute_tool(
                    tool_name=self.composio_tool.action, arguments=arguments, user_id=self.composio_tool.user_id
                )
            else:
                result = composio_service.execute_tool(tool_name=self.composio_tool.action, arguments=arguments)
        except Exception as e:
            raise NodeException(
                message=f"Error executing Composio tool '{self.composio_tool.action}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            )

        # Add result to chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield from []


class MCPNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes an MCP tool with function call output."""

    mcp_tool: MCPToolDefinition

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        try:
            mcp_service = MCPService()
            result = mcp_service.execute_tool(tool_def=self.mcp_tool, arguments=arguments)
        except Exception as e:
            raise NodeException(
                message=f"Error executing MCP tool '{self.mcp_tool.name}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            )

        # Add result to chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield from []


class ElseNode(BaseNode[ToolCallingState]):
    """Node that executes when no function conditions match."""

    class Ports(BaseNode.Ports):
        # Redefined in the create_else_node function, but defined here to resolve mypy errors
        loop_to_router = Port.on_if(
            ToolCallingState.current_prompt_output_index.less_than(ToolPromptNode.Outputs.results.length())
        )
        loop_to_prompt = Port.on_elif(ToolCallingState.current_function_calls_processed.greater_than(0))
        end = Port.on_else()

    def run(self) -> BaseNode.Outputs:
        with self.state.__quiet__():
            self.state.current_prompt_output_index += 1
        return self.Outputs()


def _hydrate_composio_tool_definition(tool_def: ComposioToolDefinition) -> FunctionDefinition:
    """Hydrate a ComposioToolDefinition with detailed information from the Composio API.

    Args:
        tool_def: The basic ComposioToolDefinition to enhance

    Returns:
        FunctionDefinition with detailed parameters and description
    """
    try:
        composio_service = ComposioService()
        tool_details = composio_service.get_tool_by_slug(tool_def.action)

        # Create a FunctionDefinition directly with proper field extraction
        return FunctionDefinition(
            name=tool_def.name,
            description=tool_details.get("description", tool_def.description),
            parameters=tool_details.get("input_parameters", {}),
        )

    except Exception as e:
        # If hydration fails (including no API key), log and return basic function definition
        logger.warning(f"Failed to enhance Composio tool '{tool_def.action}': {e}")
        return FunctionDefinition(
            name=tool_def.name,
            description=tool_def.description,
            parameters={},
        )


def hydrate_mcp_tool_definitions(server_def: MCPServer) -> List[MCPToolDefinition]:
    """Hydrate an MCPToolDefinition with detailed information from the MCP server.

    We do tool discovery on the MCP server to get the tool definitions.

    Args:
        tool_def: The basic MCPToolDefinition to enhance

    Returns:
        MCPToolDefinition with detailed parameters and description
    """
    try:
        mcp_service = MCPService()
        return mcp_service.hydrate_tool_definitions(server_def)
    except Exception as e:
        # If hydration fails, log and return original
        logger.warning(f"Failed to enhance MCP server '{server_def.name}': {e}")
        return []


def create_tool_prompt_node(
    ml_model: str,
    blocks: List[Union[PromptBlock, Dict[str, Any]]],
    functions: List[Tool],
    prompt_inputs: Optional[EntityInputsInterface],
    parameters: PromptParameters,
    max_prompt_iterations: Optional[int] = None,
) -> Type[ToolPromptNode]:
    if functions and len(functions) > 0:
        prompt_functions: List[Union[Tool, FunctionDefinition]] = []

        for function in functions:
            if isinstance(function, ComposioToolDefinition):
                # Get Composio tool details and hydrate the function definition
                enhanced_function = _hydrate_composio_tool_definition(function)
                prompt_functions.append(enhanced_function)
            elif isinstance(function, MCPServer):
                tool_functions: List[MCPToolDefinition] = hydrate_mcp_tool_definitions(function)
                for tool_function in tool_functions:
                    name = get_mcp_tool_name(tool_function)
                    prompt_functions.append(
                        FunctionDefinition(
                            name=name,
                            description=tool_function.description,
                            parameters=tool_function.parameters,
                        )
                    )
            else:
                prompt_functions.append(function)
    else:
        prompt_functions = []

    # Add a chat history block to blocks only if one doesn't already exist
    has_chat_history_block = any(
        (
            (block["block_type"] if isinstance(block, dict) else block.block_type) == "VARIABLE"
            and (
                block["input_variable"]
                if isinstance(block, dict)
                else block.input_variable if isinstance(block, VariablePromptBlock) else None
            )
            == CHAT_HISTORY_VARIABLE
        )
        for block in blocks
    )

    if not has_chat_history_block:
        blocks.append(
            VariablePromptBlock(
                block_type="VARIABLE",
                input_variable=CHAT_HISTORY_VARIABLE,
                state=None,
                cache_config=None,
            )
        )

    node_prompt_inputs = {
        **(prompt_inputs or {}),
        CHAT_HISTORY_VARIABLE: ConcatExpression[List[ChatMessage], List[ChatMessage]](
            lhs=(prompt_inputs or {}).get(CHAT_HISTORY_VARIABLE, []),
            rhs=ToolCallingState.chat_history,
        ),
    }

    node = cast(
        Type[ToolPromptNode],
        type(
            "ToolPromptNode",
            (ToolPromptNode,),
            {
                "ml_model": ml_model,
                "blocks": blocks,
                "functions": prompt_functions,  # Use converted functions for prompt layer
                "prompt_inputs": node_prompt_inputs,
                "parameters": parameters,
                "max_prompt_iterations": max_prompt_iterations,
                "__module__": __name__,
            },
        ),
    )
    return node


def create_router_node(
    functions: List[Tool],
    tool_prompt_node: Type[InlinePromptNode[ToolCallingState]],
) -> Type[RouterNode]:
    """Create a RouterNode with dynamic ports that route based on tool_prompt_node outputs."""

    if functions and len(functions) > 0:
        # Create dynamic ports and convert functions in a single loop
        Ports = type("Ports", (), {})

        # Avoid using lambda to capture function_name
        # lambda will capture the function_name by reference,
        # and if the function_name is changed, the port_condition will also change.
        def create_port_condition(fn_name):
            return Port.on_if(
                LazyReference(
                    lambda: (
                        ToolCallingState.current_prompt_output_index.less_than(
                            tool_prompt_node.Outputs.results.length()
                        )
                        & tool_prompt_node.Outputs.results[ToolCallingState.current_prompt_output_index]["type"].equals(
                            "FUNCTION_CALL"
                        )
                        & tool_prompt_node.Outputs.results[ToolCallingState.current_prompt_output_index]["value"][
                            "name"
                        ].equals(fn_name)
                    )
                )
            )

        for function in functions:
            if isinstance(function, ComposioToolDefinition):
                function_name = get_function_name(function)
                port = create_port_condition(function_name)
                setattr(Ports, function_name, port)
            elif isinstance(function, MCPServer):
                tool_functions: List[MCPToolDefinition] = hydrate_mcp_tool_definitions(function)
                for tool_function in tool_functions:
                    name = get_mcp_tool_name(tool_function)
                    port = create_port_condition(name)
                    setattr(Ports, name, port)
            else:
                function_name = get_function_name(function)
                port = create_port_condition(function_name)
                setattr(Ports, function_name, port)

        # Add the else port for when no function conditions match
        setattr(Ports, "default", Port.on_else())
    else:
        # If no functions exist, create a simple Ports class with just a default port
        Ports = type("Ports", (), {"default": Port(default=True)})

    node = cast(
        Type[RouterNode],
        type(
            "RouterNode",
            (RouterNode,),
            {
                "Ports": Ports,
                "prompt_outputs": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        ),
    )
    return node


def create_function_node(
    function: ToolBase,
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[BaseNode]:
    """
    Create a FunctionNode class for a given function.

    For workflow functions: BaseNode
    For regular functions: BaseNode with direct function call

    Args:
        function: The function to create a node for
        tool_prompt_node: The tool prompt node class
    """
    if isinstance(function, DeploymentDefinition):
        deployment = function.deployment_id or function.deployment_name
        release_tag = function.release_tag

        node = type(
            f"DynamicSubworkflowDeploymentNode_{deployment}",
            (DynamicSubworkflowDeploymentNode,),
            {
                "deployment": deployment,
                "release_tag": release_tag,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )

        return node

    elif isinstance(function, ComposioToolDefinition):
        node = type(
            f"ComposioNode_{function.name}",
            (ComposioNode,),
            {
                "composio_tool": function,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )
        return node
    elif is_workflow_class(function):
        function.is_dynamic = True
        node = type(
            f"DynamicInlineSubworkflowNode_{function.__name__}",
            (DynamicInlineSubworkflowNode,),
            {
                "subworkflow": function,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )
    else:
        # For regular functions, use FunctionNode
        node = type(
            f"FunctionNode_{function.__name__}",
            (FunctionNode,),
            {
                "function_definition": lambda self, **kwargs: function(**kwargs),  # ← Revert back to lambda
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )

    return node


def create_mcp_tool_node(
    tool_def: MCPToolDefinition,
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[BaseNode]:
    node = type(
        f"MCPNode_{tool_def.name}",
        (MCPNode,),
        {
            "mcp_tool": tool_def,
            "function_call_output": tool_prompt_node.Outputs.results,
            "__module__": __name__,
        },
    )
    return node


def create_else_node(
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[ElseNode]:
    class Ports(ElseNode.Ports):
        loop_to_router = Port.on_if(
            ToolCallingState.current_prompt_output_index.less_than(tool_prompt_node.Outputs.results.length())
        )
        loop_to_prompt = Port.on_elif(ToolCallingState.current_function_calls_processed.greater_than(0))
        end = Port.on_else()

    node = cast(
        Type[ElseNode],
        type(
            f"{tool_prompt_node.__name__}_ElseNode",
            (ElseNode,),
            {
                "Ports": Ports,
                "__module__": __name__,
            },
        ),
    )
    return node


def get_function_name(function: ToolBase) -> str:
    if isinstance(function, DeploymentDefinition):
        name = str(function.deployment_id or function.deployment_name)
        return name.replace("-", "")
    elif isinstance(function, ComposioToolDefinition):
        # model post init sets the name to the action if it's not set
        return function.name  # type: ignore[return-value]
    else:
        return snake_case(function.__name__)


def get_mcp_tool_name(tool_def: MCPToolDefinition) -> str:
    server_name = snake_case(tool_def.server.name)
    return f"{server_name}__{tool_def.name}"



from typing import List, Literal

from vellum.client.core import UniversalBaseModel
from vellum_ee.workflows.display.utils.vellum import NodeInputValuePointerRule


class NodeInputValuePointer(UniversalBaseModel):
    rules: List[NodeInputValuePointerRule]
    combinator: Literal["OR"] = "OR"


class NodeInput(UniversalBaseModel):
    id: str
    key: str
    value: NodeInputValuePointer



import pytest
from datetime import datetime
from uuid import UUID, uuid4
from typing import Type

from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import PromptDeploymentNode
from vellum_ee.workflows.display.nodes.vellum.prompt_deployment_node import BasePromptDeploymentNodeDisplay
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def _no_display_class(Node: Type[PromptDeploymentNode]):  # type: ignore
    return None


def _display_class_with_node_input_ids_by_name(Node: Type[PromptDeploymentNode]):
    class PromptDeploymentNodeDisplay(BasePromptDeploymentNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"foo": UUID("6037747a-1d35-4094-b363-4369fc92c5d4")}

    return PromptDeploymentNodeDisplay


def _display_class_with_node_input_ids_by_name_with_inputs_prefix(Node: Type[PromptDeploymentNode]):
    class PromptDeploymentNodeDisplay(BasePromptDeploymentNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"prompt_inputs.foo": UUID("6037747a-1d35-4094-b363-4369fc92c5d4")}

    return PromptDeploymentNodeDisplay


@pytest.fixture
def mock_fetch_deployment(mocker):
    # Create a mock deployment response
    mock_deployment = mocker.Mock(
        id="test-id",
        name="test-deployment",
        label="Test Deployment",
        status="ACTIVE",
        environment="DEVELOPMENT",
        created=datetime.now(),
        last_deployed_on=datetime.now(),
        last_deployed_history_item_id=str(uuid4()),
        input_variables=[],
        output_variables=[],
        description="Test deployment description",
    )

    # Patch the create_vellum_client function to return our mock client
    mocker.patch("vellum.client.resources.deployments.client.DeploymentsClient.retrieve", return_value=mock_deployment)
    return mock_deployment


@pytest.mark.parametrize(
    ["GetDisplayClass", "expected_input_id"],
    [
        (_no_display_class, "016187d6-2830-4256-a61d-e52f9bf6355e"),
        (_display_class_with_node_input_ids_by_name, "6037747a-1d35-4094-b363-4369fc92c5d4"),
        (_display_class_with_node_input_ids_by_name_with_inputs_prefix, "6037747a-1d35-4094-b363-4369fc92c5d4"),
    ],
    ids=[
        "no_display_class",
        "display_class_with_node_input_ids_by_name",
        "display_class_with_node_input_ids_by_name_with_inputs_prefix",
    ],
)
def test_serialize_node__prompt_inputs(GetDisplayClass, expected_input_id, mock_fetch_deployment):
    # GIVEN a prompt node with inputs
    class MyPromptDeploymentNode(PromptDeploymentNode):
        deployment = "DEPLOYMENT"
        prompt_inputs = {"foo": "bar"}
        ml_model_fallbacks = None

    # AND a workflow with the prompt node
    class Workflow(BaseWorkflow):
        graph = MyPromptDeploymentNode

    # AND a display class
    GetDisplayClass(MyPromptDeploymentNode)

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_prompt_node = next(
        node
        for node in serialized_workflow["workflow_raw_data"]["nodes"]
        if node["id"] == str(MyPromptDeploymentNode.__id__)
    )

    assert my_prompt_node["inputs"] == [
        {
            "id": expected_input_id,
            "key": "foo",
            "value": {
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "bar",
                        },
                    }
                ],
                "combinator": "OR",
            },
        }
    ]



from typing import TYPE_CHECKING, Any, Generic, Optional, Tuple, Type, TypeVar, Union, cast, overload

if TYPE_CHECKING:
    from vellum.workflows.expressions.accessor import AccessorExpression
    from vellum.workflows.expressions.add import AddExpression
    from vellum.workflows.expressions.and_ import AndExpression
    from vellum.workflows.expressions.begins_with import BeginsWithExpression
    from vellum.workflows.expressions.between import BetweenExpression
    from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
    from vellum.workflows.expressions.concat import ConcatExpression
    from vellum.workflows.expressions.contains import ContainsExpression
    from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
    from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
    from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
    from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression
    from vellum.workflows.expressions.ends_with import EndsWithExpression
    from vellum.workflows.expressions.equals import EqualsExpression
    from vellum.workflows.expressions.greater_than import GreaterThanExpression
    from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
    from vellum.workflows.expressions.in_ import InExpression
    from vellum.workflows.expressions.is_blank import IsBlankExpression
    from vellum.workflows.expressions.is_error import IsErrorExpression
    from vellum.workflows.expressions.is_nil import IsNilExpression
    from vellum.workflows.expressions.is_not_blank import IsNotBlankExpression
    from vellum.workflows.expressions.is_not_nil import IsNotNilExpression
    from vellum.workflows.expressions.is_not_null import IsNotNullExpression
    from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression
    from vellum.workflows.expressions.is_null import IsNullExpression
    from vellum.workflows.expressions.is_undefined import IsUndefinedExpression
    from vellum.workflows.expressions.length import LengthExpression
    from vellum.workflows.expressions.less_than import LessThanExpression
    from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
    from vellum.workflows.expressions.minus import MinusExpression
    from vellum.workflows.expressions.not_between import NotBetweenExpression
    from vellum.workflows.expressions.not_in import NotInExpression
    from vellum.workflows.expressions.or_ import OrExpression
    from vellum.workflows.expressions.parse_json import ParseJsonExpression
    from vellum.workflows.nodes.bases import BaseNode
    from vellum.workflows.state.base import BaseState

_T = TypeVar("_T", covariant=True)
_O = TypeVar("_O")
_O2 = TypeVar("_O2")


class BaseDescriptor(Generic[_T]):
    _name: str
    _types: Tuple[Type[_T], ...]
    _instance: Optional[_T]

    def __init__(self, *, name: str, types: Tuple[Type[_T], ...], instance: Optional[_T] = None) -> None:
        self._name = name
        self._types = types
        self._instance = instance

    @property
    def name(self) -> str:
        return self._name

    @property
    def types(self) -> Tuple[Type[_T], ...]:
        return self._types

    @property
    def instance(self) -> Optional[_T]:
        return self._instance

    def resolve(self, state: "BaseState") -> _T:
        raise NotImplementedError("Descriptor must implement resolve method")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._name == other._name

    def __hash__(self) -> int:
        return hash(self._name)

    def __repr__(self) -> str:
        return self._name

    @overload
    def __get__(self, instance: "BaseNode", owner: Type["BaseNode"]) -> _T: ...

    @overload
    def __get__(self, instance: Any, owner: Optional[Type]) -> "BaseDescriptor[_T]": ...

    def __get__(self, instance: Any, owner: Optional[Type]) -> Union["BaseDescriptor[_T]", _T]:
        if not instance:
            return self

        if instance.__class__.__name__ == "BaseNode":
            node = cast("BaseNode", instance)
            return self.resolve(node.state)

        return self

    # TODO: We should figure out how to remove these dynamic imports
    # https://app.shortcut.com/vellum/story/4504
    @overload
    def __or__(self, other: "BaseDescriptor[_O]") -> "OrExpression[_T, _O]": ...

    @overload
    def __or__(self, other: _O) -> "OrExpression[_T, _O]": ...

    def __or__(self, other: "Union[BaseDescriptor[_O], _O]") -> "OrExpression[_T, _O]":
        from vellum.workflows.expressions.or_ import OrExpression

        return OrExpression(lhs=self, rhs=other)

    @overload
    def __and__(self, other: "BaseDescriptor[_O]") -> "AndExpression[_T, _O]": ...

    @overload
    def __and__(self, other: _O) -> "AndExpression[_T, _O]": ...

    def __and__(self, other: "Union[BaseDescriptor[_O], _O]") -> "AndExpression[_T, _O]":
        from vellum.workflows.expressions.and_ import AndExpression

        return AndExpression(lhs=self, rhs=other)

    @overload
    def coalesce(self, other: "BaseDescriptor[_O]") -> "CoalesceExpression[_T, _O]": ...

    @overload
    def coalesce(self, other: _O) -> "CoalesceExpression[_T, _O]": ...

    def coalesce(self, other: "Union[BaseDescriptor[_O], _O]") -> "CoalesceExpression[_T, _O]":
        from vellum.workflows.expressions.coalesce_expression import CoalesceExpression

        return CoalesceExpression(lhs=self, rhs=other)

    def __getitem__(self, field: Union[str, int, "BaseDescriptor[str]", "BaseDescriptor[int]"]) -> "AccessorExpression":
        from vellum.workflows.expressions.accessor import AccessorExpression

        return AccessorExpression(base=self, field=field)

    @overload
    def equals(self, other: "BaseDescriptor[_O]") -> "EqualsExpression[_T, _O]": ...

    @overload
    def equals(self, other: _O) -> "EqualsExpression[_T, _O]": ...

    def equals(self, other: "Union[BaseDescriptor[_O], _O]") -> "EqualsExpression[_T, _O]":
        from vellum.workflows.expressions.equals import EqualsExpression

        return EqualsExpression(lhs=self, rhs=other)

    @overload
    def does_not_equal(self, other: "BaseDescriptor[_O]") -> "DoesNotEqualExpression[_T, _O]": ...

    @overload
    def does_not_equal(self, other: _O) -> "DoesNotEqualExpression[_T, _O]": ...

    def does_not_equal(self, other: "Union[BaseDescriptor[_O], _O]") -> "DoesNotEqualExpression[_T, _O]":
        from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression

        return DoesNotEqualExpression(lhs=self, rhs=other)

    @overload
    def less_than(self, other: "BaseDescriptor[_O]") -> "LessThanExpression[_T, _O]": ...

    @overload
    def less_than(self, other: _O) -> "LessThanExpression[_T, _O]": ...

    def less_than(self, other: "Union[BaseDescriptor[_O], _O]") -> "LessThanExpression[_T, _O]":
        from vellum.workflows.expressions.less_than import LessThanExpression

        return LessThanExpression(lhs=self, rhs=other)

    @overload
    def greater_than(self, other: "BaseDescriptor[_O]") -> "GreaterThanExpression[_T, _O]": ...

    @overload
    def greater_than(self, other: _O) -> "GreaterThanExpression[_T, _O]": ...

    def greater_than(self, other: "Union[BaseDescriptor[_O], _O]") -> "GreaterThanExpression[_T, _O]":
        from vellum.workflows.expressions.greater_than import GreaterThanExpression

        return GreaterThanExpression(lhs=self, rhs=other)

    @overload
    def less_than_or_equal_to(self, other: "BaseDescriptor[_O]") -> "LessThanOrEqualToExpression[_T, _O]": ...

    @overload
    def less_than_or_equal_to(self, other: _O) -> "LessThanOrEqualToExpression[_T, _O]": ...

    def less_than_or_equal_to(self, other: "Union[BaseDescriptor[_O], _O]") -> "LessThanOrEqualToExpression[_T, _O]":
        from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression

        return LessThanOrEqualToExpression(lhs=self, rhs=other)

    @overload
    def greater_than_or_equal_to(self, other: "BaseDescriptor[_O]") -> "GreaterThanOrEqualToExpression[_T, _O]": ...

    @overload
    def greater_than_or_equal_to(self, other: _O) -> "GreaterThanOrEqualToExpression[_T, _O]": ...

    def greater_than_or_equal_to(
        self, other: "Union[BaseDescriptor[_O], _O]"
    ) -> "GreaterThanOrEqualToExpression[_T, _O]":
        from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression

        return GreaterThanOrEqualToExpression(lhs=self, rhs=other)

    @overload
    def contains(self, other: "BaseDescriptor[_O]") -> "ContainsExpression[_T, _O]": ...

    @overload
    def contains(self, other: _O) -> "ContainsExpression[_T, _O]": ...

    def contains(self, other: "Union[BaseDescriptor[_O], _O]") -> "ContainsExpression[_T, _O]":
        from vellum.workflows.expressions.contains import ContainsExpression

        return ContainsExpression(lhs=self, rhs=other)

    @overload
    def begins_with(self, other: "BaseDescriptor[_O]") -> "BeginsWithExpression[_T, _O]": ...

    @overload
    def begins_with(self, other: _O) -> "BeginsWithExpression[_T, _O]": ...

    def begins_with(self, other: "Union[BaseDescriptor[_O], _O]") -> "BeginsWithExpression[_T, _O]":
        from vellum.workflows.expressions.begins_with import BeginsWithExpression

        return BeginsWithExpression(lhs=self, rhs=other)

    @overload
    def ends_with(self, other: "BaseDescriptor[_O]") -> "EndsWithExpression[_T, _O]": ...

    @overload
    def ends_with(self, other: _O) -> "EndsWithExpression[_T, _O]": ...

    def ends_with(self, other: "Union[BaseDescriptor[_O], _O]") -> "EndsWithExpression[_T, _O]":
        from vellum.workflows.expressions.ends_with import EndsWithExpression

        return EndsWithExpression(lhs=self, rhs=other)

    @overload
    def does_not_contain(self, other: "BaseDescriptor[_O]") -> "DoesNotContainExpression[_T, _O]": ...

    @overload
    def does_not_contain(self, other: _O) -> "DoesNotContainExpression[_T, _O]": ...

    def does_not_contain(self, other: "Union[BaseDescriptor[_O], _O]") -> "DoesNotContainExpression[_T, _O]":
        from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression

        return DoesNotContainExpression(lhs=self, rhs=other)

    @overload
    def does_not_begin_with(self, other: "BaseDescriptor[_O]") -> "DoesNotBeginWithExpression[_T, _O]": ...

    @overload
    def does_not_begin_with(self, other: _O) -> "DoesNotBeginWithExpression[_T, _O]": ...

    def does_not_begin_with(self, other: "Union[BaseDescriptor[_O], _O]") -> "DoesNotBeginWithExpression[_T, _O]":
        from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression

        return DoesNotBeginWithExpression(lhs=self, rhs=other)

    @overload
    def does_not_end_with(self, other: "BaseDescriptor[_O]") -> "DoesNotEndWithExpression[_T, _O]": ...

    @overload
    def does_not_end_with(self, other: _O) -> "DoesNotEndWithExpression[_T, _O]": ...

    def does_not_end_with(self, other: "Union[BaseDescriptor[_O], _O]") -> "DoesNotEndWithExpression[_T, _O]":
        from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression

        return DoesNotEndWithExpression(lhs=self, rhs=other)

    def is_null(self) -> "IsNullExpression":
        from vellum.workflows.expressions.is_null import IsNullExpression

        return IsNullExpression(expression=self)

    def is_not_null(self) -> "IsNotNullExpression":
        from vellum.workflows.expressions.is_not_null import IsNotNullExpression

        return IsNotNullExpression(expression=self)

    def is_nil(self) -> "IsNilExpression":
        from vellum.workflows.expressions.is_nil import IsNilExpression

        return IsNilExpression(expression=self)

    def is_not_nil(self) -> "IsNotNilExpression":
        from vellum.workflows.expressions.is_not_nil import IsNotNilExpression

        return IsNotNilExpression(expression=self)

    def is_undefined(self) -> "IsUndefinedExpression":
        from vellum.workflows.expressions.is_undefined import IsUndefinedExpression

        return IsUndefinedExpression(expression=self)

    def is_not_undefined(self) -> "IsNotUndefinedExpression":
        from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression

        return IsNotUndefinedExpression(expression=self)

    @overload
    def in_(self, other: "BaseDescriptor[_O]") -> "InExpression[_T, _O]": ...

    @overload
    def in_(self, other: _O) -> "InExpression[_T, _O]": ...

    def in_(self, other: "Union[BaseDescriptor[_O], _O]") -> "InExpression[_T, _O]":
        from vellum.workflows.expressions.in_ import InExpression

        return InExpression(lhs=self, rhs=other)

    @overload
    def not_in(self, other: "BaseDescriptor[_O]") -> "NotInExpression[_T, _O]": ...

    @overload
    def not_in(self, other: _O) -> "NotInExpression[_T, _O]": ...

    def not_in(self, other: "Union[BaseDescriptor[_O], _O]") -> "NotInExpression[_T, _O]":
        from vellum.workflows.expressions.not_in import NotInExpression

        return NotInExpression(lhs=self, rhs=other)

    @overload
    def between(self, start: "BaseDescriptor[_O]", end: "BaseDescriptor[_O2]") -> "BetweenExpression[_T, _O, _O2]": ...

    @overload
    def between(self, start: _O, end: _O2) -> "BetweenExpression[_T, _O, _O2]": ...

    def between(
        self, start: "Union[BaseDescriptor[_O], _O]", end: "Union[BaseDescriptor[_O2], _O2]"
    ) -> "BetweenExpression[_T, _O, _O2]":
        from vellum.workflows.expressions.between import BetweenExpression

        return BetweenExpression(value=self, start=start, end=end)

    @overload
    def not_between(
        self, start: "BaseDescriptor[_O]", end: "BaseDescriptor[_O2]"
    ) -> "NotBetweenExpression[_T, _O, _O2]": ...

    @overload
    def not_between(self, start: _O, end: _O2) -> "NotBetweenExpression[_T, _O, _O2]": ...

    def not_between(
        self, start: "Union[BaseDescriptor[_O], _O]", end: "Union[BaseDescriptor[_O2], _O2]"
    ) -> "NotBetweenExpression[_T, _O, _O2]":
        from vellum.workflows.expressions.not_between import NotBetweenExpression

        return NotBetweenExpression(value=self, start=start, end=end)

    def is_blank(self) -> "IsBlankExpression":
        from vellum.workflows.expressions.is_blank import IsBlankExpression

        return IsBlankExpression(expression=self)

    def is_not_blank(self) -> "IsNotBlankExpression":
        from vellum.workflows.expressions.is_not_blank import IsNotBlankExpression

        return IsNotBlankExpression(expression=self)

    def parse_json(self) -> "ParseJsonExpression[_T]":
        from vellum.workflows.expressions.parse_json import ParseJsonExpression

        return ParseJsonExpression(expression=self)

    def is_error(self) -> "IsErrorExpression":
        from vellum.workflows.expressions.is_error import IsErrorExpression

        return IsErrorExpression(expression=self)

    @overload
    def concat(self, other: "BaseDescriptor[_O]") -> "ConcatExpression[_T, _O]": ...

    @overload
    def concat(self, other: _O) -> "ConcatExpression[_T, _O]": ...

    def concat(self, other: "Union[BaseDescriptor[_O], _O]") -> "ConcatExpression[_T, _O]":
        from vellum.workflows.expressions.concat import ConcatExpression

        return ConcatExpression(lhs=self, rhs=other)

    def length(self) -> "LengthExpression[_T]":
        from vellum.workflows.expressions.length import LengthExpression

        return LengthExpression(expression=self)

    @overload
    def add(self, other: "BaseDescriptor[_O]") -> "AddExpression[_T, _O]": ...

    @overload
    def add(self, other: _O) -> "AddExpression[_T, _O]": ...

    def add(self, other: "Union[BaseDescriptor[_O], _O]") -> "AddExpression[_T, _O]":
        from vellum.workflows.expressions.add import AddExpression

        return AddExpression(lhs=self, rhs=other)

    @overload
    def minus(self, other: "BaseDescriptor[_O]") -> "MinusExpression[_T, _O]": ...

    @overload
    def minus(self, other: _O) -> "MinusExpression[_T, _O]": ...

    def minus(self, other: "Union[BaseDescriptor[_O], _O]") -> "MinusExpression[_T, _O]":
        from vellum.workflows.expressions.minus import MinusExpression

        return MinusExpression(lhs=self, rhs=other)



import json
from typing import Any, Union

from jinja2 import Undefined
from jinja2.utils import htmlsafe_json_dumps

from vellum.workflows.state.encoder import DefaultStateEncoder


def is_valid_json_string(value: Union[str, bytes]) -> bool:
    """Determines whether the given value is a valid JSON string."""

    try:
        json.loads(value)
    except ValueError:
        return False
    return True


def replace(s: Any, old: Any, new: Any) -> str:
    def encode_to_str(obj: Any) -> str:
        """Encode an object for template rendering using DefaultStateEncoder."""
        try:
            if isinstance(obj, str):
                return obj
            return json.dumps(obj, cls=DefaultStateEncoder)
        except TypeError:
            return str(obj)

    if old == "":
        return encode_to_str(s)

    s_str = encode_to_str(s)
    old_str = encode_to_str(old)
    new_str = encode_to_str(new)
    return s_str.replace(old_str, new_str)


def safe_tojson(value: Any) -> str:
    if isinstance(value, Undefined):
        return ""
    return htmlsafe_json_dumps(value, cls=DefaultStateEncoder)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from vellum.workflows.events.workflow import WorkflowEvent
from vellum.workflows.state.base import BaseState

# To protect against circular imports
if TYPE_CHECKING:
    from vellum.workflows.state.context import WorkflowContext


class BaseWorkflowEmitter(ABC):
    def __init__(self):
        self._context: Optional["WorkflowContext"] = None

    def register_context(self, context: "WorkflowContext") -> None:
        """
        Register the workflow context with this emitter.

        Args:
            context: The workflow context containing shared resources like vellum_client.
        """
        self._context = context

    @abstractmethod
    def emit_event(self, event: WorkflowEvent) -> None:
        pass

    @abstractmethod
    def snapshot_state(self, state: BaseState) -> None:
        pass



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

from .audio_prompt_block import AudioPromptBlock
from .document_prompt_block import DocumentPromptBlock
from .function_call_prompt_block import FunctionCallPromptBlock
from .image_prompt_block import ImagePromptBlock
from .jinja_prompt_block import JinjaPromptBlock
from .rich_text_prompt_block import RichTextPromptBlock
from .variable_prompt_block import VariablePromptBlock
from .video_prompt_block import VideoPromptBlock

if typing.TYPE_CHECKING:
    from .chat_message_prompt_block import ChatMessagePromptBlock
PromptBlock = typing.Union[
    JinjaPromptBlock,
    "ChatMessagePromptBlock",
    VariablePromptBlock,
    RichTextPromptBlock,
    FunctionCallPromptBlock,
    AudioPromptBlock,
    VideoPromptBlock,
    ImagePromptBlock,
    DocumentPromptBlock,
]



from typing import TYPE_CHECKING, Generic, Optional, Tuple, Type, TypeVar, cast

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.types.generics import import_workflow_class

if TYPE_CHECKING:
    from vellum.workflows.inputs.base import BaseInputs
    from vellum.workflows.state.base import BaseState

_InputType = TypeVar("_InputType")


class WorkflowInputReference(BaseDescriptor[_InputType], Generic[_InputType]):

    def __init__(
        self,
        *,
        name: str,
        types: Tuple[Type[_InputType], ...],
        instance: Optional[_InputType],
        inputs_class: Type["BaseInputs"],
    ) -> None:
        super().__init__(name=name, types=types, instance=instance)
        self._inputs_class = inputs_class

    @property
    def inputs_class(self) -> Type["BaseInputs"]:
        return self._inputs_class

    def resolve(self, state: "BaseState") -> _InputType:
        if hasattr(state.meta.workflow_inputs, self._name) and (
            state.meta.workflow_definition == self._inputs_class.__parent_class__
            or not issubclass(self._inputs_class.__parent_class__, import_workflow_class())
        ):
            return cast(_InputType, getattr(state.meta.workflow_inputs, self._name))

        if state.meta.parent:
            return self.resolve(state.meta.parent)

        if type(None) in self.types:
            return cast(_InputType, None)

        raise NodeException(f"Missing required Workflow input: {self._name}", code=WorkflowErrorCode.INVALID_INPUTS)

    def __repr__(self) -> str:
        return f"{self._inputs_class.__qualname__}.{self.name}"



from vellum.workflows.edges.edge import Edge
from vellum.workflows.graph.graph import Graph
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.ports.port import Port


def test_graph__from_node():
    # GIVEN a node
    class MyNode(BaseNode):
        pass

    # WHEN we create a graph from the node
    graph = Graph.from_node(MyNode)

    # THEN the graph has the node as the entrypoint
    assert set(graph.entrypoints) == {MyNode}

    # AND one node
    assert len(list(graph.nodes)) == 1

    # AND no edges
    assert len(list(graph.edges)) == 0


def test_graph__from_edge():
    # GIVEN an edge
    class SourceNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    edge = Edge(SourceNode.Ports.default, TargetNode)

    # WHEN we create a graph from the edge
    graph = Graph.from_edge(edge)

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND one edge
    assert len(list(graph.edges)) == 1


def test_graph__node_to_node():
    # GIVEN two nodes
    class SourceNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from the source node to the target node
    graph = SourceNode >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND one edge
    assert len(list(graph.edges)) == 1


def test_graph__port_to_node():
    # GIVEN a two nodes, where the source node has a port
    class SourceNode(BaseNode):
        class Ports(BaseNode.Ports):
            output = Port.on_else()

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from the source port to the target node
    graph = SourceNode.Ports.output >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND one edge
    assert len(list(graph.edges)) == 1


def test_graph__graph_to_node():
    # GIVEN three nodes
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from the source node to the target node via the middle node
    graph = SourceNode >> MiddleNode >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__edgeless_graph_to_node():
    # GIVEN two nodes
    class SourceNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # AND a graph of the first node
    subgraph = Graph.from_node(SourceNode)

    # WHEN we create a graph from the source graph to the target node
    graph = subgraph >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND one edge
    assert len(list(graph.edges)) == 1

    # AND the first node's port has reference to the edge
    assert set(SourceNode.Ports.default.edges) == set(graph.edges)


def test_graph__edgeless_graph_to_graph():
    # GIVEN two nodes
    class SourceNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # AND a graph of the first node
    subgraph = Graph.from_node(SourceNode)

    # AND a graph of the second node
    target_subgraph = Graph.from_node(TargetNode)

    # WHEN we create a graph from the source graph to the target node
    graph = subgraph >> target_subgraph

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND one edge
    assert len(list(graph.edges)) == 1

    # AND the first node's port has reference to the edges
    assert set(SourceNode.Ports.default.edges) == set(graph.edges)


def test_graph__graph_to_edgeless_graph():
    # GIVEN three nodes
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # AND a graph of the first two nodes
    subgraph = SourceNode >> MiddleNode

    # AND a graph of the third node
    target_subgraph = Graph.from_node(TargetNode)

    # WHEN we create a graph from the source graph to the target node
    graph = subgraph >> target_subgraph

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__node_to_graph():
    # GIVEN three nodes
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # AND a graph of the last two nodes
    target_subgraph = MiddleNode >> TargetNode

    # WHEN we create a graph from the source node to the target graph
    graph = SourceNode >> target_subgraph

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__repeated_edge():
    # GIVEN two nodes
    class SourceNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from the source node to the target node twice
    graph = SourceNode >> TargetNode >> SourceNode >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND two nodes
    assert len(list(graph.nodes)) == 2

    # AND two edges
    assert len(list(graph.edges)) == 2

    # AND the first node's port has reference to just one edge
    assert len(list(SourceNode.Ports.default.edges)) == 1


def test_graph__node_to_set():
    # GIVEN three nodes, one with ports
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        class Ports(BaseNode.Ports):
            top = Port.on_if(SourceNode.Execution.count.less_than(1))
            bottom = Port.on_else()

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph with a set
    graph = SourceNode >> {
        MiddleNode.Ports.top >> SourceNode,
        MiddleNode.Ports.bottom >> TargetNode,
    }

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 3


def test_graph__graph_to_set():
    # GIVEN four nodes, one with ports
    class SourceNode(BaseNode):
        pass

    class SecondNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        class Ports(BaseNode.Ports):
            top = Port.on_if(SourceNode.Execution.count.less_than(1))
            bottom = Port.on_else()

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph between a graph and a set
    graph = (
        SourceNode
        >> SecondNode
        >> {
            MiddleNode.Ports.top >> SourceNode,
            MiddleNode.Ports.bottom >> TargetNode,
        }
    )

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 4

    # AND two edges
    assert len(list(graph.edges)) == 4


def test_graph__graph_set_to_set():
    # GIVEN five nodes
    class SourceNode(BaseNode):
        pass

    class TopNode(BaseNode):
        pass

    class BottomNode(BaseNode):
        pass

    class ConditionalNode(BaseNode):
        class Ports(BaseNode.Ports):
            top = Port.on_if(SourceNode.Execution.count.less_than(1))
            bottom = Port.on_else()

    class EndNode(BaseNode):
        pass

    # WHEN we create a graph that draws an edge between sets
    graph = (
        SourceNode
        >> {
            TopNode,
            BottomNode,
        }
        >> {
            ConditionalNode.Ports.top >> SourceNode,
            ConditionalNode.Ports.bottom >> EndNode,
        }
    )

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 5

    # AND two edges
    assert len(list(graph.edges)) == 6


def test_graph__port_to_graph():
    # GIVEN three nodes where the first node has a port
    class SourceNode(BaseNode):
        class Ports(BaseNode.Ports):
            custom = Port.on_else()

    class MiddleNode(BaseNode):
        pass

    class EndNode(BaseNode):
        pass

    # WHEN we create a graph from the port to a subgraph
    graph = SourceNode.Ports.custom >> (MiddleNode >> EndNode)

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__port_to_set():
    # GIVEN three nodes where the first node has a port
    class SourceNode(BaseNode):
        class Ports(BaseNode.Ports):
            custom = Port.on_else()

    class TopNode(BaseNode):
        pass

    class BottomNode(BaseNode):
        pass

    # WHEN we create a graph from the port to the set
    graph = SourceNode.Ports.custom >> {
        TopNode,
        BottomNode,
    }

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__set_to_node():
    # GIVEN three nodes
    class TopNode(BaseNode):
        pass

    class BottomNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from a set to a node
    graph = {
        TopNode,
        BottomNode,
    } >> TargetNode

    # THEN the graph has both the top node and the bottom node as the entrypoints
    assert set(graph.entrypoints) == {TopNode, BottomNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__node_to_port():
    # GIVEN two nodes, one with a port
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        class Ports(BaseNode.Ports):
            custom = Port.on_else()

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from the source node to the target node
    graph = SourceNode >> MiddleNode.Ports.custom >> TargetNode

    # THEN the graph has the source node as the entrypoint
    assert set(graph.entrypoints) == {SourceNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__set_to_graph():
    # GIVEN three nodes
    class SourceNode(BaseNode):
        pass

    class MiddleNode(BaseNode):
        pass

    class TargetNode(BaseNode):
        pass

    # WHEN we create a graph from a set to a graph
    graph: Graph = {SourceNode, MiddleNode} >> Graph.from_node(TargetNode)

    # THEN the graph has the source node and middle node as the entrypoints
    assert set(graph.entrypoints) == {SourceNode, MiddleNode}

    # AND three nodes
    assert len(list(graph.nodes)) == 3

    # AND two edges
    assert len(list(graph.edges)) == 2


def test_graph__str_simple_linear():
    # GIVEN a simple linear graph: A -> B -> C
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    graph = NodeA >> NodeB >> NodeC

    # WHEN we convert the graph to string
    result = str(graph)

    # THEN it shows the linear flow structure
    expected_lines = ["Graph:", "  └─ NodeA", "     └─ NodeB", "        └─ NodeC"]
    assert result == "\n".join(expected_lines)


def test_graph__str_with_branching():
    # GIVEN a graph with branching: A -> {B, C}
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    graph = NodeA >> {NodeB, NodeC}

    # WHEN we convert the graph to string
    result = str(graph)

    # THEN it shows the branching structure
    # Note: The order might vary due to set ordering, so we check for the structure
    lines = result.split("\n")
    assert lines[0] == "Graph:"
    assert lines[1] == "  └─ NodeA"

    # Should have two branches (order may vary)
    branch_lines = [line.strip() for line in lines[2:] if line.strip()]
    assert len(branch_lines) == 2
    assert any("NodeB" in line for line in branch_lines)
    assert any("NodeC" in line for line in branch_lines)
    assert all(line.startswith("├─ ") or line.startswith("└─ ") for line in branch_lines)


def test_graph__str_with_loop():
    # GIVEN a graph with a loop: A -> B -> A (loop back)
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    # Create the loop manually using edges
    edge1 = Edge(NodeA.Ports.default, NodeB)
    edge2 = Edge(NodeB.Ports.default, NodeA)

    graph = Graph(entrypoints={NodeA.Ports.default}, edges=[edge1, edge2], terminals={NodeA.Ports.default})

    # WHEN we convert the graph to string
    result = str(graph)

    # THEN it shows the loop with cycle detection
    expected_lines = ["Graph:", "  └─ NodeA", "     └─ NodeB", "        └─ NodeA ⟲ (loops back)"]
    assert result == "\n".join(expected_lines)


def test_graph__str_empty_graph():
    # GIVEN an empty graph
    graph = Graph(entrypoints=set(), edges=[], terminals=set())

    # WHEN we convert the graph to string
    result = str(graph)

    # THEN it shows empty graph message
    assert result == "Graph(empty)"


def test_graph__str_single_node():
    # GIVEN a graph with just one node and no edges
    class SingleNode(BaseNode):
        pass

    graph = Graph.from_node(SingleNode)

    # WHEN we convert the graph to string
    result = str(graph)

    # THEN it shows the single node
    assert "SingleNode.default" in result
    assert "Graph:" in result



import pytest
from dataclasses import dataclass
import os
import shutil
import tempfile
from uuid import uuid4
from typing import Any, Callable, Dict, Generator

import tomli_w


@dataclass
class MockModuleResult:
    temp_dir: str
    module: str
    set_pyproject_toml: Callable[[Dict[str, Any]], None]
    workflow_sandbox_id: str


@pytest.fixture
def mock_module(request) -> Generator[MockModuleResult, None, None]:
    current_dir = os.getcwd()
    temp_dir = tempfile.mkdtemp()
    os.chdir(temp_dir)

    # Use the test name to create a unique module path
    module = f"examples.mock.{request.node.name}"
    workflow_sandbox_id = str(uuid4())

    def set_pyproject_toml(vellum_config: Dict[str, Any]) -> None:
        pyproject_toml_path = os.path.join(temp_dir, "pyproject.toml")
        with open(pyproject_toml_path, "wb") as f:
            tomli_w.dump(
                {"tool": {"vellum": vellum_config}},
                f,
            )

    set_pyproject_toml(
        {
            "workflows": [
                {
                    "module": module,
                    "workflow_sandbox_id": workflow_sandbox_id,
                }
            ]
        }
    )

    with open(os.path.join(temp_dir, ".env"), "w") as f:
        f.write("VELLUM_API_KEY=abcdef123456")

    yield MockModuleResult(
        temp_dir=temp_dir,
        module=module,
        set_pyproject_toml=set_pyproject_toml,
        workflow_sandbox_id=workflow_sandbox_id,
    )

    os.chdir(current_dir)
    shutil.rmtree(temp_dir)


@pytest.fixture
def info_log_level(monkeypatch):
    """Set log level to INFO for tests that request this fixture"""
    monkeypatch.setenv("LOG_LEVEL", "INFO")



from uuid import uuid4
from typing import Any, Iterator, List

from vellum import (
    ExecutePromptEvent,
    FulfilledExecutePromptEvent,
    InitiatedExecutePromptEvent,
    PromptOutput,
    StringVellumValue,
)
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import PromptDeploymentNode
from vellum.workflows.state import BaseState
from vellum.workflows.state.base import StateMeta


def test_text_prompt_deployment_node__basic(vellum_client):
    """Confirm that TextPromptDeploymentNodes output the expected text and results when run."""

    # GIVEN a node that subclasses TextPromptDeploymentNode
    class Inputs(BaseInputs):
        input: str

    class State(BaseState):
        pass

    class MyPromptDeploymentNode(PromptDeploymentNode):
        deployment = "my-deployment"

    # AND a known response from invoking a deployed prompt
    expected_outputs: List[PromptOutput] = [
        StringVellumValue(value="Hello, world!"),
    ]

    def generate_prompt_events(*args: Any, **kwargs: Any) -> Iterator[ExecutePromptEvent]:
        execution_id = str(uuid4())
        events: List[ExecutePromptEvent] = [
            InitiatedExecutePromptEvent(execution_id=execution_id),
            FulfilledExecutePromptEvent(
                execution_id=execution_id,
                outputs=expected_outputs,
            ),
        ]
        yield from events

    vellum_client.execute_prompt_stream.side_effect = generate_prompt_events

    # WHEN the node is run
    node = MyPromptDeploymentNode(
        state=State(
            meta=StateMeta(workflow_inputs=Inputs(input="Say something.")),
        )
    )
    outputs = [o for o in node.run()]

    # THEN the node should have produced the outputs we expect
    results_output = outputs[0]
    assert results_output.name == "results"
    assert results_output.value == expected_outputs

    text_output = outputs[1]
    assert text_output.name == "text"
    assert text_output.value == "Hello, world!"

    # AND we should have made the expected call to stream the prompt execution
    vellum_client.execute_prompt_stream.assert_called_once()
    _, call_kwargs = vellum_client.execute_prompt_stream.call_args
    exec_ctx = call_kwargs["request_options"]["additional_body_parameters"]["execution_context"]
    assert exec_ctx["parent_context"] is not None
    assert exec_ctx["parent_context"]["type"] == "EXTERNAL"



import importlib
from importlib.machinery import ModuleSpec
import re
import sys
from typing import Optional


class VirtualFileLoader(importlib.abc.Loader):
    def __init__(self, files: dict[str, str], namespace: str):
        self.files = files
        self.namespace = namespace

    def create_module(self, spec: ModuleSpec):
        """
        We started with cpython/Lib/importlib/_bootstrap.py::FrozenImporter::create_module here

        https://github.com/python/cpython/blob/053c285f6b41f92fbdd1d4ff0c959cceefacd7cd/Lib/importlib/_bootstrap.py#L1160C1-L1169C22

        and reduced our needs to just updating the __file__ attribute directly.
        """
        module = type(sys)(spec.name)
        module.__file__ = spec.origin
        return module

    def exec_module(self, module):
        module_info = self._resolve_module(module.__spec__.name)

        if module_info:
            file_path, code = module_info
            compiled = compile(code, file_path, "exec")
            exec(compiled, module.__dict__)

    def get_source(self, fullname):
        """
        `inspect` module uses this method to get the source code of a module.
        """

        module_info = self._resolve_module(fullname)
        if module_info:
            return module_info[1]

        return None

    def _resolve_module(self, fullname: str) -> Optional[tuple[str, str]]:
        file_path = self._get_file_path(fullname)
        code = self._get_code(file_path)

        if code is not None:
            return file_path, code

        if not file_path.endswith("__init__.py"):
            file_path = re.sub(r"\.py$", "/__init__.py", file_path)
            code = self._get_code(file_path)

            if code is not None:
                return file_path, code

        return None

    def _get_file_path(self, fullname):
        return f"{fullname.replace('.', '/')}.py"

    def _get_code(self, file_path):
        file_key_name = re.sub(r"^" + re.escape(self.namespace) + r"/", "", file_path)
        return self.files.get(file_key_name)


class VirtualFileFinder(importlib.abc.MetaPathFinder, importlib.abc.Loader):
    def __init__(self, files: dict[str, str], namespace: str):
        self.loader = VirtualFileLoader(files, namespace)

    def find_spec(self, fullname: str, path, target=None):
        module_info = self.loader._resolve_module(fullname)
        if module_info:
            file_path, _ = module_info
            is_package = file_path.endswith("__init__.py")
            return importlib.machinery.ModuleSpec(
                fullname,
                self.loader,
                origin=file_path,
                is_package=is_package,
            )

        return None



# This file was auto-generated by Fern from our API Definition.

import typing

from .workflow_result_event_output_data_array import WorkflowResultEventOutputDataArray
from .workflow_result_event_output_data_chat_history import WorkflowResultEventOutputDataChatHistory
from .workflow_result_event_output_data_error import WorkflowResultEventOutputDataError
from .workflow_result_event_output_data_function_call import WorkflowResultEventOutputDataFunctionCall
from .workflow_result_event_output_data_json import WorkflowResultEventOutputDataJson
from .workflow_result_event_output_data_number import WorkflowResultEventOutputDataNumber
from .workflow_result_event_output_data_search_results import WorkflowResultEventOutputDataSearchResults
from .workflow_result_event_output_data_string import WorkflowResultEventOutputDataString

WorkflowResultEventOutputData = typing.Union[
    WorkflowResultEventOutputDataString,
    WorkflowResultEventOutputDataNumber,
    WorkflowResultEventOutputDataJson,
    WorkflowResultEventOutputDataChatHistory,
    WorkflowResultEventOutputDataSearchResults,
    WorkflowResultEventOutputDataArray,
    WorkflowResultEventOutputDataFunctionCall,
    WorkflowResultEventOutputDataError,
]



import logging
import os


def load_logger() -> logging.Logger:
    logger = logging.getLogger(__package__)
    logger.setLevel(os.getenv("LOG_LEVEL", logging.DEBUG))

    # Add a stream handler so that we see logs in the console when running tests
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
    logger.addHandler(handler)

    return logger



from abc import abstractmethod
from typing import ClassVar, Generator, Generic, Iterator, List, Optional, Union

from vellum import AdHocExecutePromptEvent, ExecutePromptEvent, PromptOutput
from vellum.client.core import RequestOptions
from vellum.client.core.api_error import ApiError
from vellum.workflows.errors.types import WorkflowErrorCode, vellum_error_to_workflow_error
from vellum.workflows.events.node import NodeExecutionStreamingEvent
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutput, BaseOutputs
from vellum.workflows.references.output import OutputReference
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior
from vellum.workflows.types.generics import StateType


class BasePromptNode(BaseNode[StateType], Generic[StateType]):
    # Inputs that are passed to the Prompt
    prompt_inputs: ClassVar[Optional[EntityInputsInterface]] = None

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseOutputs):
        results: List[PromptOutput]

    @abstractmethod
    def _get_prompt_event_stream(self) -> Union[Iterator[AdHocExecutePromptEvent], Iterator[ExecutePromptEvent]]:
        pass

    def _validate(self) -> None:
        pass

    def run(self) -> Iterator[BaseOutput]:
        outputs = yield from self._process_prompt_event_stream()
        if outputs is None:
            raise NodeException(
                message="Expected to receive outputs from Prompt",
                code=WorkflowErrorCode.INTERNAL_ERROR,
            )

    def _process_prompt_event_stream(self) -> Generator[BaseOutput, None, Optional[List[PromptOutput]]]:
        self._validate()
        try:
            prompt_event_stream = self._get_prompt_event_stream()
        except ApiError as e:
            self._handle_api_error(e)

        # We don't use the INITIATED event anyway, so we can just skip it
        # and use the exception handling to catch other api level errors
        try:
            next(prompt_event_stream)
        except ApiError as e:
            self._handle_api_error(e)

        outputs: Optional[List[PromptOutput]] = None
        for event in prompt_event_stream:
            if event.state == "INITIATED":
                continue
            elif event.state == "STREAMING":
                yield BaseOutput(name="results", delta=event.output.value)
            elif event.state == "FULFILLED":
                outputs = event.outputs
                yield BaseOutput(name="results", value=event.outputs)
            elif event.state == "REJECTED":
                workflow_error = vellum_error_to_workflow_error(event.error)
                raise NodeException.of(workflow_error)

        return outputs

    def _handle_api_error(self, e: ApiError):
        if e.status_code and e.status_code == 403 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Provider credentials is missing or unavailable"),
                code=WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE,
            )

        elif e.status_code and e.status_code >= 400 and e.status_code < 500 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Failed to execute Prompt"),
                code=WorkflowErrorCode.INVALID_INPUTS,
            ) from e

        raise NodeException(
            message="Failed to execute Prompt",
            code=WorkflowErrorCode.INTERNAL_ERROR,
        ) from e

    def __directly_emit_workflow_output__(
        self,
        event: NodeExecutionStreamingEvent,
        workflow_output_descriptor: OutputReference,
    ) -> bool:
        if event.output.name != "results":
            return False

        if not isinstance(event.output.delta, str) and not event.output.is_initiated:
            return False

        target_nodes = [e.to_node for port in self.Ports for e in port.edges if e.to_node.__simulates_workflow_output__]
        target_node_output = next(
            (
                o
                for target_node in target_nodes
                for o in target_node.Outputs
                if o == workflow_output_descriptor.instance
            ),
            None,
        )
        if not target_node_output:
            return False

        return True



# This file was auto-generated by Fern from our API Definition.

import typing

from .workflow_request_chat_history_input_request import WorkflowRequestChatHistoryInputRequest
from .workflow_request_json_input_request import WorkflowRequestJsonInputRequest
from .workflow_request_number_input_request import WorkflowRequestNumberInputRequest
from .workflow_request_string_input_request import WorkflowRequestStringInputRequest

WorkflowRequestInputRequest = typing.Union[
    WorkflowRequestStringInputRequest,
    WorkflowRequestJsonInputRequest,
    WorkflowRequestChatHistoryInputRequest,
    WorkflowRequestNumberInputRequest,
]
